import unittest


class TestEPIFM(unittest.TestCase):

    def setUp(self):
        self.radial_cutoff = 1000.0e-9
        self.radial_resolution = 1.0e-9

    def tearDown(self):
        pass

    def test1(self):
        import scopyon._epifm

    def test2(self):
        print('Testing TRITC ...')
        import numpy
        from scopyon._epifm import PointSpreadingFunction
        psf = PointSpreadingFunction(psf_radial_cutoff=self.radial_cutoff, psf_radial_width=None, psf_depth_cutoff=1000.0e-9, fluorophore_type="Tetramethylrhodamine(TRITC)", psf_wavelength=5.78e-07)

        depth = 0.0
        radial = numpy.arange(0.0, self.radial_cutoff, self.radial_resolution, dtype=float)
        psf_r = psf.get_distribution(radial, depth)
        self.assertIs(type(psf_r), numpy.ndarray)
        self.assertEqual(psf_r.ndim, 1)
        self.assertEqual(psf_r.size, radial.size)
        self.assertTrue((psf_r >= 0.0).all())

        tot_r = numpy.sum(2 * numpy.pi * radial * psf_r) * self.radial_resolution
        print(f'Integral of radial distribution = {tot_r}')

        psf_cart = psf.radial_to_cartesian(radial, psf_r, self.radial_cutoff, self.radial_resolution)
        tot_cart = psf_cart.sum() * (self.radial_resolution * self.radial_resolution)
        print(f'Integral of cartesian distribution = {tot_cart}')

        camera = numpy.zeros((512, 512))
        pixel_length = 4.444444444444444e-08
        psf.overlay_signal_(camera, psf_cart, numpy.zeros(3, dtype=float), pixel_length, self.radial_resolution, 1.0)
        # tot_camera = camera.sum() * (self.radial_resolution * self.radial_resolution)
        tot_camera = camera.sum()
        print(f'Integral of detected = {tot_camera}')

    def test3(self):
        print('Testing Gaussian ...')
        import numpy
        from scopyon._epifm import PointSpreadingFunction
        psf = PointSpreadingFunction(psf_radial_cutoff=self.radial_cutoff, psf_radial_width=1.0e-7, psf_depth_cutoff=1000.0e-9, fluorophore_type="Gaussian", psf_wavelength=6.0e-7)

        depth = 0.0
        radial = numpy.arange(0.0, self.radial_cutoff, self.radial_resolution, dtype=float)
        psf_r = psf.get_distribution(radial, depth)
        self.assertIs(type(psf_r), numpy.ndarray)
        self.assertEqual(psf_r.ndim, 1)
        self.assertEqual(psf_r.size, radial.size)
        self.assertTrue((psf_r >= 0.0).all())

        tot_r = numpy.sum(2 * numpy.pi * radial * psf_r) * self.radial_resolution
        print(f'Integral of radial distribution = {tot_r}')

        psf_cart = psf.radial_to_cartesian(radial, psf_r, self.radial_cutoff, self.radial_resolution)
        tot_cart = psf_cart.sum() * (self.radial_resolution * self.radial_resolution)
        print(f'Integral of cartesian distribution = {tot_cart}')

        camera = numpy.zeros((512, 512))
        pixel_length = 4.444444444444444e-08
        psf.overlay_signal_(camera, psf_cart, numpy.zeros(3, dtype=float), pixel_length, self.radial_resolution, 1.0)
        # tot_camera = camera.sum() * (self.radial_resolution * self.radial_resolution)
        tot_camera = camera.sum()
        print(f'Integral of detected = {tot_camera}')


if __name__ == '__main__':
    unittest.main()



import scopyon
import numpy.random


config = scopyon.DefaultConfiguration()
config.default.detector.exposure_time = 33.0e-3
pixel_length = config.default.detector.pixel_length / config.default.magnification
L_2 = config.default.detector.image_size[0] * pixel_length * 0.5

rng = numpy.random.RandomState(123)
N = 100
inputs = rng.uniform(-L_2, +L_2, size=(N, 2))

img = scopyon.form_image(inputs, config=config, rng=rng)
img.save("tirf_000.png")



import collections.abc
import contextlib
import io
import pkgutil
import warnings

import numpy

from scopyon.constants import Q_  # pint

from logging import getLogger
_log = getLogger(__name__)

__all__ = ["Configuration", "DefaultConfiguration"]


def dict_merge(dct, merge_dct):
    for k, v in merge_dct.items():
        if isinstance(dct.get(k), dict) and isinstance(v, collections.Mapping):
            dict_merge(dct[k], v)
        else:
            dct[k] = v

class Configuration(collections.abc.Mapping):
    """

    Note:
        Requires `yaml` and `pint`.
    """

    def __init__(self, filename=None, yaml=None):
        if filename is not None:
            assert yaml is None
            import yaml
            try:
                from yaml import CLoader as Loader
            except ImportError:
                from yaml import Loader
            with open(filename) as f:
                self.__yaml = yaml.load(f.read(), Loader=Loader)
        elif yaml is not None:
            self.__yaml = yaml
        else:
            self.__yaml = None

    def __repr__(self):
        import yaml
        try:
            from yaml import CDumper as Dumper
        except ImportError:
            from yaml import Dumper
        return yaml.dump(self.__yaml, default_flow_style=False, Dumper=Dumper)

    def update(self, conf):
        if isinstance(conf, Configuration):
            dict_merge(self.__yaml, conf.yaml)
        elif isinstance(conf, dict):
            dict_merge(self.__yaml, conf)
        else:
            import yaml
            try:
                from yaml import CLoader as Loader
            except ImportError:
                from yaml import Loader
            dict_merge(self.__yaml, yaml.load(conf, Loader=Loader))

    @property
    def yaml(self):
        return self.__yaml

    def get(self, key, defaultobj=None):
        return self.__yaml.get(key, defaultobj)

    def __getitem__(self, key):
        return getattr(self, key)
        # value = self.__yaml[key]
        # if isinstance(value, dict):
        #     assert 'value' in value
        #     return value['value']
        # return value

    def __len__(self):
        return len(self.__yaml)

    def __iter__(self):
        return (key for key, value in self.__yaml.items() if not isinstance(value, dict) or 'value' in value)

    def __getattr__(self, key):
        if key not in self.__yaml:
            raise KeyError(f"'{key}'")
        value = self.__yaml[key]
        if isinstance(value, dict):
            if 'value' not in value:
                return Configuration(yaml=value)
            elif value['value'] is not None and 'units' in value:
                #XXX: value['value'] could be None (null).
                given = Q_(value['value'], value['units'])
                v = given.to_base_units()
                if given.units != v.units:
                    warnings.warn(f"Unit conversion in '{key}' from [{given.units:~}] to [{v.units:~}]")
                return v.magnitude
            else:
                return value['value']
        return value

    def __setattr__(self, key, value):
        if key.startswith('_'):
            object.__setattr__(self, key, value)
            return
        if key not in self.__yaml:
            raise KeyError(f"'{key}'")
        if isinstance(value, dict):
            raise TypeError(f"The given value for '{key}' has wrong type: {value}")

        original = self.__yaml[key]
        if isinstance(original, dict):
            if 'value' not in original:
                #TODO: value might be Configuration.
                raise ValueError(f"Cannot update '{key}'.")
        if isinstance(value, Q_):
            if isinstance(original, dict) and 'units' in original:
                other = Q_(original['value'], original['units'])
                if not value.check(other):
                    from pint.errors import DimensionalityError
                    raise DimensionalityError(
                        value.units, other.units, value.dimensionality, other.dimensionality)
            self.__yaml[key] = dict(value=value.magnitude, units='{:~}'.format(value.units))
        else:
            self.__yaml[key] = value  #XXX: clear 'units'

class DefaultConfiguration(Configuration):
    """

    Default configuration. All the settings are available in package data, `scopyon.yaml`.
    See https://github.com/ecell/scopyon/blob/master/scopyon/scopyon.yaml
    """

    def __init__(self):
        import yaml
        try:
            from yaml import CLoader as Loader
        except ImportError:
            from yaml import Loader
        with contextlib.closing(io.StringIO(pkgutil.get_data("scopyon", "scopyon.yaml").decode())) as f:
            data = yaml.load(f.read(), Loader=Loader)

        Configuration.__init__(self, yaml=data)



import numpy

__all__ = ['show']


def __get_shape(shape):
    import matplotlib.patches as patches
    row, column, sigma, c = shape['x'], shape['y'], shape['sigma'], shape['color']
    x, y = column, row  # imshow
    return patches.Rectangle((x - sigma, y - sigma), 2 * sigma, 2 * sigma, color=c, linewidth=1, fill=False)

def show(img, shapes=None):
    """Show an image.

    Args:
        img (ndarray): An image data to be shown.
        shapes (list, optional): A list of shapes.
            shape is a dictionary consisting of `x` (row), `y` (column), `sigma` and `color`.
            `sigma` is a half size of the box (square).
    """
    if not isinstance(img, numpy.ndarray):
        raise TypeError("'img' must be an Image or array.")

    import matplotlib.pyplot as plt
    _, ax = plt.subplots()
    if img.ndim == 2:
        plt.imshow(img, interpolation='none', cmap='gray')
    elif img.ndim == 3:
        plt.imshow(img, interpolation='none')
    else:
        raise ValueError("'img' has wrong dimension.")
    if shapes is not None:
        for shape in shapes:
            ax.add_patch(__get_shape(shape))
    plt.show()



import scopyon
import numpy.random


config = scopyon.DefaultConfiguration()
config.update("""
default:
    magnification: 360
    detector:
        exposure_time: 0.033
    effects:
        photo_bleaching:
            switch: true
            half_life: 2.5
""")
pixel_length = config.default.detector.pixel_length / config.default.magnification
L_2 = config.default.detector.image_size[0] * pixel_length * 0.5

rng = numpy.random.RandomState(123)
num_frames = 30
N = 1000
D = 0.1e-12  # m ** 2 / s
dt = config.default.detector.exposure_time
t = numpy.arange(0, (num_frames + 1) * dt, dt)
inputs = scopyon.sample_inputs(t, N=N, lower=-L_2, upper=+L_2, ndim=2, D=D, rng=rng)

img = list(scopyon.generate_images(inputs, num_frames=num_frames, config=config, rng=rng))
scopyon.Video.save('video.mp4', img)



import os.path
import csv
import contextlib
import io
import pkgutil
from collections import defaultdict

from logging import getLogger
_log = getLogger(__name__)


def isfloat(val):
    """Check if the given string can be converted to float or not.
    This method is fast when returning True

    Args:
        val (str): The input

    Returns:
        bool: Return True if the input can be converted to float

    """
    try:
        float(val)
        return True
    except ValueError:
        return False

def read_catalog(data_path):
    package_name = 'scopyon'
    data = pkgutil.get_data(package_name, data_path).decode()

    catalog_data = defaultdict(list)
    with contextlib.closing(io.StringIO(data)) as fin:
        for _ in range(5):
            line = fin.readline()
            line = line.rstrip()
            _log.debug('     {}'.format(line))

        reader = csv.reader(fin)

        row = next(reader)
        if len(row) != 1 or isfloat(row[0]):
            raise RuntimeError('A catalog in invalid format was given [{}]'.format(filename))
        key = row[0]

        for row in reader:
            if len(row) == 1 and not isfloat(row[0]):
                key = row[0]
            elif len(row) != 0:
                catalog_data[key].append(row)
    return catalog_data

def read_fluorophore_catalog(fluorophore_type):
    catalog_data = read_catalog('catalog/fluorophore/{}.csv'.format(fluorophore_type))
    return catalog_data['Excitation'], catalog_data['Emission']

def read_dichroic_catalog(dm):
    return read_catalog('catalog/dichroic/{}.csv'.format(dm))['wavedataset']

def read_excitation_catalog(excitation):
    return read_catalog('catalog/excitation/{}.csv'.format(excitation))['wavedataset']

def read_emission_catalog(emission):
    filename = os.path.join(os.path.abspath(os.path.dirname(__file__)), )
    return read_catalog('catalog/emission/{}.csv'.format(emission))['wavedataset']



import numpy

#XXX: Never import plotly here. Do inside each function.

__all__ = ['show']


def __get_shape(shape):
    row, column, sigma, c = shape['x'], shape['y'], shape['sigma'], shape['color']
    x, y = column, row  # imshow
    return dict(type='rect', x0=x - sigma, y0=y - sigma, x1=x + sigma, y1=y + sigma, line=dict(color=c))

def show(img, shapes=None):
    """Show an image.

    Note:
        Requires `plotly.express`.

    Args:
        img (ndarray): An image data to be shown.
        shapes (list, optional): A list of shapes.
            shape is a dictionary consisting of `x` (row), `y` (column), `sigma` and `color`.
            `sigma` is a half size of the box (square).
    """
    if not isinstance(img, numpy.ndarray):
        raise TypeError("'img' must be an Image or array.")

    if shapes is not None:
        shapes = [__get_shape(shape) for shape in shapes]

    import plotly.express as px
    if img.ndim == 2:
        fig = px.imshow(img, color_continuous_scale='gray')
    elif img.ndim == 3:
        fig = px.imshow(img)
    else:
        raise ValueError("'img' has wrong dimension.")
    if shapes is not None:
        fig.update_layout(shapes=shapes)
    fig.update_layout(xaxis=dict(range=[0, img.shape[0]]), yaxis=dict(range=[0, img.shape[1]]))
    fig.show()



import copy
import math
import warnings
import os
import numbers
import functools
from multiprocessing import Pool

import numpy

import scipy.special
import scipy.integrate
from scipy.special import j0, i1e
from scipy.interpolate import interp1d
from scipy.ndimage import map_coordinates

from . import io
from .image import Image
from . import constants

from logging import getLogger
_log = getLogger(__name__)


# def _polar2cartesian_coordinates(r, t, x, y):
#     X, Y = numpy.meshgrid(x, y)
#     new_r = numpy.sqrt(X * X + Y * Y)
#     new_t = numpy.arctan2(X, Y)
# 
#     ir = interp1d(r, numpy.arange(len(r)), bounds_error=False)
#     it = interp1d(t, numpy.arange(len(t)))
# 
#     new_ir = ir(new_r.ravel())
#     new_it = it(new_t.ravel())
#     new_ir[new_r.ravel() > r.max()] = len(r) - 1
#     new_ir[new_r.ravel() < r.min()] = 0
# 
#     return numpy.array([new_ir, new_it])
# 
# def _polar2cartesian(grid, coordinates, shape):
#     r = shape[0] - 1
#     psf_cart = numpy.empty([2 * r + 1, 2 * r + 1])
#     psf_cart[r: , r: ] = map_coordinates(grid, coordinates, order=0).reshape(shape)
#     psf_cart[r: , : r] = psf_cart[r: , : r: -1]
#     psf_cart[: r, : ] = psf_cart[: r: -1, : ]
#     return psf_cart

class PointSpreadingFunction:

    def __init__(self, psf_radial_cutoff, psf_radial_width, psf_depth_cutoff, fluorophore_type, psf_wavelength):
        # fluorophore
        self.fluorophore_type = fluorophore_type
        self.psf_wave_length = psf_wavelength
        self.psf_radial_width = psf_radial_width  #XXX: Only required for Gaussian

        self.__fluorophore_psf = {}
        self.__cutoff_radial = psf_radial_cutoff
        self.__cutoff_depth = psf_depth_cutoff
        self.__resolution_depth = 1e-9
        self.__resolution_radial = 1e-9

        if self.fluorophore_type == 'Gaussian':
            if self.psf_radial_width is None:
                raise ValueError(
                        'fluorophore.radial_width must be given for Gaussian type fluorophore.')
            self.get_distribution = self.get_gaussian_distribution
            #XXX: `partial` is not picklable.
            # self.get_distribution = functools.partial(
            #         self.__get_gaussian_distribution, radial_width=self.psf_radial_width)
        else:
            self.get_distribution = self.get_born_wolf_distribution
            #XXX: `partial` is not picklable.
            # self.get_distribution = functools.partial(
            #         self.__get_born_wolf_distribution_old, wave_length=self.psf_wave_length)

    def get(self, depth):
        depth = abs(depth)
        if depth < self.__cutoff_depth + self.__resolution_depth:
            assert depth >= 0
            key = int(depth / self.__resolution_depth)
            depth = key * self.__resolution_depth
        else:
            key = -1
            depth = self.__cutoff_depth

        if key not in self.__fluorophore_psf:
            self.__fluorophore_psf[key] = self.__get(depth)
        return self.__fluorophore_psf[key]

    def __get(self, depth):
        radial = numpy.arange(0.0, self.__cutoff_radial, self.__resolution_radial, dtype=float)
        r = radial / self.__resolution_radial
        radial_cutoff = self.__cutoff_radial / self.__resolution_radial
        psf_r = self.get_distribution(radial, depth)
        psf_cart = self.radial_to_cartesian(radial, psf_r, self.__cutoff_radial, self.__resolution_radial)
        return psf_cart

    @staticmethod
    def radial_to_cartesian(radial, radial_distribution, radial_cutoff, radial_resolution):
        # r = radial / radial_resolution
        # radial_cutoff = radial_cutoff / radial_resolution

        # theta = numpy.linspace(0, 90, 91)  #XXX: theta resolution
        # z = numpy.linspace(0, radial_cutoff, len(r))
        # y = numpy.linspace(0, radial_cutoff, len(r))

        # coordinates = _polar2cartesian_coordinates(r, theta, z, y)
        # theta = numpy.ones_like(theta)
        # polar_distribution = numpy.array(list(map(lambda x: theta * x, radial_distribution)))
        # return _polar2cartesian(polar_distribution, coordinates, (len(z), len(y)))

        assert isinstance(radial, numpy.ndarray) and radial.ndim == 1
        assert isinstance(radial_distribution, numpy.ndarray) and radial.ndim == 1
        assert radial.size == radial_distribution.size
        assert isinstance(radial_cutoff, numbers.Real)
        assert isinstance(radial_resolution, numbers.Real)

        shape = (2 * (radial.size  - 1) + 1, 2 * (radial.size  - 1) + 1)
        X, Y = numpy.meshgrid(numpy.arange(shape[0]), numpy.arange(shape[1]))
        X = (X.ravel() - (radial.size - 1)) * radial_resolution
        Y = (Y.ravel() - (radial.size - 1)) * radial_resolution
        R = numpy.sqrt(X ** 2 + Y ** 2)
        R[R > radial.max()] = radial.max()  #XXX
        interp_func = interp1d(radial, radial_distribution)
        P = interp_func(R)
        return P.reshape(shape)

    def get_gaussian_distribution(self, radial, depth):
        psf = self.__get_gaussian_distribution(radial, depth, self.psf_radial_width)
        return psf

    @staticmethod
    def __get_gaussian_distribution(r, _z, radial_width):
        return numpy.exp(-0.5 * (r / radial_width) ** 2) / (2 * numpy.pi * radial_width * radial_width)

        # #XXX: normalization means I in __get_normalization.
        # # For normalization
        # # norm = list(map(lambda x: True if x > 1e-4 else False, I))
        # norm = (normalization > 1e-4)
        # # Ir = sum(list(map(lambda x: x * numpy.exp(-0.5 * (r / self.psf_width[0]) ** 2), norm)))
        # # Iz = sum(list(map(lambda x: x * numpy.exp(-0.5 * (z / self.psf_width[1]) ** 2), norm)))
        # Ir = norm * numpy.exp(-0.5 * numpy.power(r / width[0], 2))
        # Iz = norm * numpy.exp(-0.5 * numpy.power(r / width[1], 2))
        # return numpy.array(list(map(lambda x: Ir * x, Iz)))

    def get_born_wolf_distribution(self, radial, depth):
        # psf = self.__get_born_wolf_distribution(radial, depth, self.psf_wave_length)
        psf = self.__get_born_wolf_distribution_old(radial, depth, self.psf_wave_length)
        # NA = 1.4  # self.objective_NA
        # k = 2.0 * numpy.pi / self.psf_wave_length
        # alpha = k * NA
        # unit_area = 1e-9 ** 2
        # psf *= 1.0 / (alpha * alpha / numpy.pi * unit_area)  #TODO: Unknown normalization factor
        # print("Normalization:", 1.0 / (alpha * alpha / numpy.pi * unit_area))
        return psf

    @staticmethod
    def __get_born_wolf_distribution(r, z, wave_length):
        assert isinstance(r, numpy.ndarray) and r.ndim == 1
        assert isinstance(z, numbers.Real) and z >= 0.0
        assert isinstance(wave_length, numbers.Real) and wave_length > 0

        # set Numerical Appature
        NA = 1.4  # self.objective_NA

        # set alpha and psi
        k = 2.0 * numpy.pi / wave_length
        alpha = k * NA
        psi = 0.5 * alpha * z * NA

        psf = numpy.zeros(r.size, dtype=r.dtype)
        for i in range(r.size):
            alpha_r = alpha * r[i]
            I_real, _ = scipy.integrate.quad(lambda rho: j0(alpha_r * rho) * +numpy.cos(psi * rho * rho) * rho, 0, 1.0)
            I_imag, _ = scipy.integrate.quad(lambda rho: j0(alpha_r * rho) * -numpy.sin(psi * rho * rho) * rho, 0, 1.0)
            psf[i] = I_real ** 2 + I_imag ** 2
        psf *= alpha * alpha / numpy.pi  #XXX: normalization
        return psf

    @staticmethod
    def __get_born_wolf_distribution_old(r, z, wave_length):
        # set Numerical Appature
        NA = 1.4  # self.objective_NA

        # set alpha and gamma consts
        k = 2.0 * numpy.pi / wave_length
        alpha = k * NA
        gamma = k * numpy.power(NA / 2, 2)

        # set rho parameters
        N = 100
        drho = 1.0 / N
        # rho = numpy.array([(i + 1) * drho for i in range(N)])
        rho = numpy.arange(1, N + 1) * drho
        # print(f'r.shape = {r.shape}')
        # print(f'rho.shape = {rho.shape}')

        J0 = numpy.array(list(map(lambda x: j0(x * alpha * rho), r)))
        # print(f'J0.shape = {J0.shape}')
        Y  = numpy.exp(-2 * 1.j * z * gamma * rho * rho) * rho * drho
        # print(f'Y.shape = {Y.shape}')
        I  = Y * J0
        # print(f'I.shape = {I.shape}')
        I_sum = I.sum(axis=1)
        # print(f'I_sum.shape = {I_sum.shape}')

        # set PSF
        psf = numpy.array(list(map(lambda x: abs(x) ** 2, I_sum)))
        # print(f'psf.shape = {psf.shape}')
        # print(f'psf.sum = {psf.sum()}')

        psf *= alpha * alpha / numpy.pi  #XXX: normalization
        return psf

    def overlay_signal(self, expected, p_i, pixel_length, normalization=1.0):
        # fluorophore axial position
        if normalization <= 0.0:
            return
        depth = p_i[0]
        signal = self.get(depth)
        self.overlay_signal_(expected, signal, p_i, pixel_length, self.__resolution_radial, normalization)

    @staticmethod
    def overlay_signal_(expected, signal, p_i, pixel_length, signal_resolution, normalization):
        _, xi, yi = p_i
        Nw_pixel, Nh_pixel = expected.shape
        # signal_resolution = 1.0e-9  # m  #=> self.__cutoff_radial / (len(r) - 1) ???
        signal_width = signal_resolution * (signal.shape[0] - 1)
        signal_height = signal_resolution * (signal.shape[1] - 1)
        expected_width = Nw_pixel * pixel_length  # Not `(Nw_pixel - 1) * pixel_length`
        expected_height = Nh_pixel * pixel_length  # Not `(Nh_pixel - 1) * pixel_length`

        imin = math.floor((expected_width * 0.5 + xi - signal_width * 0.5) / pixel_length)
        imax = math.ceil((expected_width * 0.5 + xi + signal_width * 0.5) / pixel_length)
        iarray = numpy.arange(max(0, imin), min(Nw_pixel, imax) + 1)
        left = (iarray * pixel_length - (expected_width * 0.5 + xi - signal_width * 0.5)) / signal_resolution
        left = numpy.ceil(left).astype(int)
        # assert left[1] > 0
        # if imin >= 0:
        #     assert left[0] <= 0
        #     left[0] = 0
        # else:
        #     assert left[0] > 0
        # assert left[-2] < signal.shape[0]
        # if imax <= Nw_pixel:
        #     assert left[-1] >= signal.shape[0]
        #     left[-1] = signal.shape[0]
        # else:
        #     assert left[-1] < signal.shape[0]
        left[0] = max(left[0], 0)
        left[-1] = min(left[-1], signal.shape[0])

        jmin = math.floor((expected_height * 0.5 + yi - signal_height * 0.5) / pixel_length)
        jmax = math.ceil((expected_height * 0.5 + yi + signal_height * 0.5) / pixel_length)
        jarray = numpy.arange(max(0, jmin), min(Nh_pixel, jmax) + 1)
        top = (jarray * pixel_length - (expected_height * 0.5 + yi - signal_height * 0.5)) / signal_resolution
        top = numpy.ceil(top).astype(int)
        # assert top[1] > 0
        # if jmin >= 0:
        #     assert top[0] <= 0
        #     top[0] = 0
        # else:
        #     assert top[0] > 0
        # assert top[-2] < signal.shape[1]
        # if jmax <= Nw_pixel:
        #     assert top[-1] >= signal.shape[1]
        #     top[-1] = signal.shape[1]
        # else:
        #     assert top[-1] < signal.shape[1]
        top[0] = max(top[0], 0)
        top[-1] = min(top[-1], signal.shape[1])

        unit_area = signal_resolution * signal_resolution
        for i, i0, i1 in zip(iarray, left[: -1], left[1: ]):
            for j, j0, j1 in zip(jarray, top[: -1], top[1: ]):
                photons = signal[i0: i1, j0: j1].sum() * unit_area
                if photons > 0:
                    expected[i, j] += photons * normalization

        # # particle position
        # _, yi, zi = p_i
        # dy = (yi - signal_resolution * signal.shape[0] / 2 + pixel_length * (Nw_pixel - 1) / 2) / pixel_length
        # dz = (zi - signal_resolution * signal.shape[1] / 2 + pixel_length * (Nh_pixel - 1) / 2) / pixel_length
        # imin = math.ceil(dy)
        # imax = math.ceil(signal.shape[0] / pixel_ratio + dy) - 1
        # jmin = math.ceil(dz)
        # jmax = math.ceil(signal.shape[1] / pixel_ratio + dz) - 1
        # imin = max(imin, 0)
        # imax = min(imax, Nw_pixel)
        # jmin = max(jmin, 0)
        # jmax = min(jmax, Nh_pixel)

        # iarray = numpy.arange(imin, imax + 1)
        # ibottom = (iarray[: -1] - dy) * pixel_ratio
        # ibottom = numpy.maximum(numpy.floor(ibottom).astype(int), 0)
        # itop = (iarray[1: ] - dy) * pixel_ratio
        # itop = numpy.minimum(numpy.floor(itop).astype(int), signal.shape[0])
        # irange = numpy.vstack((iarray[: -1], ibottom, itop)).T
        # irange = irange[irange[:, 2] > irange[:, 1]]

        # jarray = numpy.arange(jmin, jmax)
        # jbottom = (jarray[: -1] - dz) * pixel_ratio
        # jbottom = numpy.maximum(numpy.floor(jbottom).astype(int), 0)
        # jtop = (jarray[1: ] - dz) * pixel_ratio
        # jtop = numpy.minimum(numpy.floor(jtop).astype(int), signal.shape[1])
        # jrange = numpy.vstack((jarray[: -1], jbottom, jtop)).T
        # jrange = jrange[jrange[:, 2] > jrange[:, 1]]

        # unit_area = signal_resolution * signal_resolution
        # for i, i0, i1 in irange:
        #     for j, j0, j1 in jrange:
        #         photons = signal[i0: i1, j0: j1].sum() * unit_area
        #         if photons > 0:
        #             expected[i, j] += photons * normalization

def cylindrical_coordinate(p_i, p_0):
    p_i = numpy.array(p_i)  # Make a copy
    p_0 = numpy.asarray(p_0)

    displacement = p_i - p_0
    radial = numpy.sqrt(displacement[1] ** 2 + displacement[2] ** 2)
    depth = displacement[0]
    return p_i, radial, depth

class CMOS:

    @staticmethod
    def get_noise(shape, rng, dtype=numpy.float64):
        ## get detector noise (photoelectrons)
        noise_data = numpy.loadtxt(
            os.path.join(os.path.abspath(os.path.dirname(__file__)), 'catalog/detector/RNDist_F40.csv'),
            delimiter=',')
        Nr_cmos = noise_data[: , 0]
        p_noise = noise_data[: , 1]
        p_nsum  = p_noise.sum()
        noise = numpy.zeros(shape, dtype=dtype)
        for i in range(shape[0]):
            for j in range(shape[1]):
                ## get detector noise (photoelectrons)
                noise[i, j]  = rng.choice(Nr_cmos, None, p=p_noise / p_nsum)
        return noise

    @staticmethod
    def get_signal(expected, rng=None, processes=None):
        signal = numpy.zeros(expected.shape, dtype=expected.dtype)
        for i in range(expected.shape[0]):
            for j in range(expected.shape[1]):
                signal[i, j] = rng.poisson(expected[i, j], None)
        return signal

class EMCCD:

    @staticmethod
    def get_noise(shape, readout_noise, rng, dtype=numpy.float64):
        ## get detector noise (photoelectrons)
        if readout_noise <= 0:
            return numpy.zeros(shape, dtype=dtype)
        return rng.normal(0, readout_noise, shape)

    @staticmethod
    def probability(S, E, a):
        # numpy.sqrt(a * E / S) * numpy.exp(-a * S - E + 2 * numpy.sqrt(a * E * S)) * i1e(2 * numpy.sqrt(a * E * S))
        X = a * S
        Y = 2 * numpy.sqrt(E * X)
        return (1.0 / Y * numpy.exp(-X + Y) * i1e(Y)) # * (2 * a * E * numpy.exp(-E))

    @staticmethod
    def probability_distribution(expected, emgain):
        ## set probability distributions
        sigma = numpy.sqrt(expected) * 5 + 10
        s_min = max(0, emgain * int(expected - sigma))
        s_max = emgain * int(expected + sigma)
        # s = numpy.array([k for k in range(s_min, s_max)])
        S = numpy.arange(s_min, s_max)

        a = 1.0 / emgain
        if S[0] > 0:
            p_signal = EMCCD.probability(S, expected, a)
        else:
            # assert S[1] > 0
            p_signal = numpy.zeros(shape=(len(S), ))
            # p_signal[0] = numpy.exp(-expected)
            p_signal[0] = 1.0 / (2 * a * expected)
            p_signal[1: ] = EMCCD.probability(S[1: ], expected, a)

        p_signal /= p_signal.sum()
        return S, p_signal

    @staticmethod
    def draw_signal(expected, emgain, rng=None):
        if expected <= 0:
            return 0
        s, p_signal = EMCCD.probability_distribution(expected, emgain)
        ## get signal (photoelectrons)
        signal = (rng or numpy.random).choice(s, None, p=p_signal)
        return signal

    @staticmethod
    def get_signal(expected, emgain, rng=None, processes=None):
        if processes is not None and processes > 1:
            with Pool(processes) as pool:
                signal = pool.map(functools.partial(EMCCD.get_signal, rng=None, emgain=emgain), numpy.array_split(expected.flatten(), processes))
            signal = numpy.hstack(signal).reshape(expected.shape)
        else:
            # epected could be one-dimensional. See above.
            shape = expected.shape
            expected = expected.flatten()
            signal = numpy.zeros(expected.size, dtype=expected.dtype)
            for i in range(signal.size):
                signal[i] = EMCCD.draw_signal(expected[i], emgain, rng)
            signal = signal.reshape(shape)
        return signal

class CCD:

    @staticmethod
    def get_noise(shape, readout_noise, rng, dtype=numpy.float64):
        ## get detector noise (photoelectrons)
        if readout_noise <= 0:
            return numpy.zeros(shape, dtype=dtype)
        return rng.normal(0, readout_noise, shape)

    @staticmethod
    def get_signal(expected, rng=None, processes=None):
        signal = numpy.zeros(expected.shape, dtype=expected.dtype)
        for i in range(expected.shape[0]):
            for j in range(expected.shape[1]):
                signal[i, j] = rng.poisson(expected[i, j], None)
        return signal

class PhysicalEffectConfigs:
    '''
    Physical effects setting class

        Fluorescence
        Photo-bleaching
        Photo-blinking
    '''

    def __init__(self, config):
        self.initialize(config)

    def initialize(self, config):
        self.fluorescence_bleach = []
        self.fluorescence_budget = []
        self.fluorescence_state  = []

        self.set_background(**config.background)
        self.set_fluorescence(**config.fluorescence)
        self.set_photobleaching(**config.photo_bleaching)
        # self.set_photoactivation(**config.photo_activation)
        # self.set_photoblinking(**config.photo_blinking)

    def set_background(self, mean=None, switch=True):
        self.background_switch = switch
        self.background_mean = mean
        _log.info('--- Background: ')
        _log.info('    Mean = {} photons'.format(self.background_mean))

    def set_fluorescence(self, quantum_yield=None, abs_coefficient=None):
        self.quantum_yield = quantum_yield
        self.abs_coefficient = abs_coefficient
        _log.info('--- Fluorescence: ')
        _log.info('    Quantum Yield =  {}'.format(self.quantum_yield))
        _log.info('    Abs. Coefficient =  {} 1/(M cm)'.format(self.abs_coefficient))
        _log.info('    Abs. Cross-section =  {} cm^2'.format((numpy.log(10) * self.abs_coefficient * 0.1 / constants.N_A) * 1e+4))

    def set_photobleaching(self, half_life=None, switch=True):
        self.photobleaching_switch = switch
        self.photobleaching_half_life = half_life
        _log.info('--- Photobleaching: ')
        _log.info('    Photobleaching half life  =  {}'.format(self.photobleaching_half_life))

    # def set_photoactivation(self, turn_on_ratio=None, activation_yield=None, frac_preactivation=None, switch=True):
    #     self.photoactivation_switch = switch
    #     self.photoactivation_turn_on_ratio = turn_on_ratio
    #     self.photoactivation_activation_yield = activation_yield
    #     self.photoactivation_frac_preactivation = frac_preactivation
    #     _log.info('--- Photoactivation: ')
    #     _log.info('    Turn-on Ratio  =  {}'.format(self.photoactivation_turn_on_ratio))
    #     _log.info('    Effective Ratio  =  {}'.format(self.photoactivation_activation_yield * self.photoactivation_turn_on_ratio / (1 + self.photoactivation_frac_preactivation * self.photoactivation_turn_on_ratio)))
    #     _log.info('    Reaction Yield =  {}'.format(self.photoactivation_activation_yield))
    #     _log.info('    Fraction of Preactivation =  {}'.format(self.photoactivation_frac_preactivation))

    # def set_photoblinking(self, t0_on=None, a_on=None, t0_off=None, a_off=None, switch=True):
    #     self.photoblinking_switch = switch
    #     self.photoblinking_t0_on = t0_on
    #     self.photoblinking_a_on = a_on
    #     self.photoblinking_t0_off = t0_off
    #     self.photoblinking_a_off = a_off
    #     _log.info('--- Photo-blinking: ')
    #     _log.info('    (ON)  t0 =  {} sec'.format(self.photoblinking_t0_on))
    #     _log.info('    (ON)  a  =  {}'.format(self.photoblinking_a_on))
    #     _log.info('    (OFF) t0 =  {} sec'.format(self.photoblinking_t0_off))
    #     _log.info('    (OFF) a  =  {}'.format(self.photoblinking_a_off))

    # def get_prob_bleach(self, tau, dt):
    #     # set the photobleaching-time
    #     tau0  = self.photobleaching_tau0
    #     alpha = self.photobleaching_alpha

    #     # set probability
    #     prob = self.prob_levy(tau, tau0, alpha)
    #     norm = prob.sum()
    #     p_bleach = prob/norm

    #     return p_bleach

    # def get_prob_blink(self, tau_on, tau_off, dt):
    #     # time scale
    #     t0_on  = self.photoblinking_t0_on
    #     a_on   = self.photoblinking_a_on
    #     t0_off = self.photoblinking_t0_off
    #     a_off  = self.photoblinking_a_off

    #     # ON-state
    #     prob_on = self.prob_levy(tau_on, t0_on, a_on)
    #     norm_on = prob_on.sum()
    #     p_blink_on = prob_on/norm_on

    #     # OFF-state
    #     prob_off = self.prob_levy(tau_off, t0_off, a_off)
    #     norm_off = prob_off.sum()
    #     p_blink_off = prob_off/norm_off

    #     return p_blink_on, p_blink_off

    # def get_photobleaching_property(self, dt, n_emit0, rng):
    #     # set the photobleaching-time
    #     tau0  = self.photobleaching_tau0

    #     # set photon budget
    #     photon0 = (tau0/dt)*n_emit0

    #     tau_bleach = numpy.array([j*dt + tau0 for j in range(int(1e+7))])
    #     p_bleach = self.get_prob_bleach(tau_bleach, dt)

    #     # get photon budget and photobleaching-time
    #     tau = rng.choice(tau_bleach, 1, p=p_bleach)
    #     budget = (tau/tau0)*photon0

    #     return tau, budget

    # def set_photophysics_4palm(self, start, end, dt, f, F, N_part, rng):
    #     ##### PALM Configuration
    #     NNN = int(1e+7)

    #     # set the photobleaching-time
    #     tau0  = self.photobleaching_tau0
    #     alpha = self.photobleaching_alpha

    #     # set photon budget
    #     # photon0 = (tau0/dt)*N_emit0
    #     photon0 = (tau0/dt)*1.0

    #     tau_bleach = numpy.array([j*dt+tau0 for j in range(NNN)])
    #     prob = self.prob_levy(tau_bleach, tau0, alpha)
    #     norm = prob.sum()
    #     p_bleach = prob/norm

    #     # set photoblinking (ON/OFF) probability density function (PDF)
    #     if (self.photoblinking_switch == True):

    #         # time scale
    #         t0_on  = self.photoblinking_t0_on
    #         a_on   = self.photoblinking_a_on
    #         t0_off = self.photoblinking_t0_off
    #         a_off  = self.photoblinking_a_off

    #         tau_on  = numpy.array([j*dt+t0_on  for j in range(NNN)])
    #         tau_off = numpy.array([j*dt+t0_off for j in range(NNN)])

    #         # ON-state
    #         prob_on = self.prob_levy(tau_on, t0_on, a_on)
    #         norm_on = prob_on.sum()
    #         p_on = prob_on/norm_on

    #         # OFF-state
    #         prob_off = self.prob_levy(tau_off, t0_off, a_off)
    #         norm_off = prob_off.sum()
    #         p_off = prob_off/norm_off

    #     # sequences
    #     budget = []
    #     state_act = []

    #     # get random number
    #     # numpy.random.seed()

    #     # get photon budget and photobleaching-time
    #     tau = rng.choice(tau_bleach, N_part, p=p_bleach)

    #     for i in range(N_part):

    #         # get photon budget and photobleaching-time
    #         photons = (tau[i]/tau0)*photon0

    #         N = int((end - start)/dt)

    #         time  = numpy.array([j*dt + start for j in range(N)])
    #         state = numpy.zeros(shape=(N))

    #         #### Photoactivation: overall activation yeild
    #         p = self.photoactivation_activation_yield
    #         # q = self.photoactivation_frac_preactivation

    #         r = rng.uniform(0, 1, N/F*f)
    #         s = (r < p).astype('int')

    #         index = numpy.abs(s - 1).argmin()
    #         t0 = (index/f*F + numpy.remainder(index, f))*dt + start

    #         t1 = t0 + tau[i]

    #         N0 = (numpy.abs(time - t0)).argmin()
    #         N1 = (numpy.abs(time - t1)).argmin()
    #         state[N0:N1] = numpy.ones(shape=(N1-N0))

    #         if (self.photoblinking_switch == True):

    #             # ON-state
    #             t_on  = rng.choice(tau_on,  100, p=p_on)
    #             # OFF-state
    #             t_off = rng.choice(tau_off, 100, p=p_off)

    #             k = (numpy.abs(numpy.cumsum(t_on) - tau[i])).argmin()

    #             # merge t_on/off arrays
    #             t_blink = numpy.array([t_on[0:k], t_off[0:k]])
    #             t = numpy.reshape(t_blink, 2*k, order='F')

    #             i_on  = N0

    #             for j in range(k):

    #                 f_on  = i_on  + int(t[j]/dt)
    #                 i_off = f_on
    #                 f_off = i_off + int(t[j+1]/dt)

    #                 state[i_on:f_on]   = 1
    #                 state[i_off:f_off] = 0

    #                 i_on = f_off

    #         # set fluorescence state (without photoblinking)
    #         budget.append(photons)
    #         state_act.append(state)

    #     self.fluorescence_bleach = numpy.array(tau)
    #     self.fluorescence_budget = numpy.array(budget)
    #     self.fluorescence_state  = numpy.array(state_act)

    # def set_photophysics_4lscm(self, start, end, dt, N_part, rng):
    #     ##### LSCM Configuration

    #     NNN = int(1e+7)

    #     # set the photobleaching-time
    #     tau0  = self.photobleaching_tau0
    #     alpha = self.photobleaching_alpha

    #     # set photon budget
    #     photon0 = (tau0/dt)*1.0

    #     tau_bleach = numpy.array([j*dt+tau0 for j in range(NNN)])
    #     prob = self.prob_levy(tau_bleach, tau0, alpha)
    #     norm = prob.sum()
    #     p_bleach = prob/norm

    #     # set photoblinking (ON/OFF) probability density function (PDF)
    #     if (self.photoblinking_switch == True):

    #         # time scale
    #         t0_on  = self.photoblinking_t0_on
    #         a_on   = self.photoblinking_a_on
    #         t0_off = self.photoblinking_t0_off
    #         a_off  = self.photoblinking_a_off

    #         tau_on  = numpy.array([j*dt+t0_on  for j in range(NNN)])
    #         tau_off = numpy.array([j*dt+t0_off for j in range(NNN)])

    #         # ON-state
    #         prob_on = self.prob_levy(tau_on, t0_on, a_on)
    #         norm_on = prob_on.sum()
    #         p_on = prob_on/norm_on

    #         # OFF-state
    #         prob_off = self.prob_levy(tau_off, t0_off, a_off)
    #         norm_off = prob_off.sum()
    #         p_off = prob_off/norm_off

    #     # sequences
    #     budget = []
    #     state_act = []

    #     # get random number
    #     # numpy.random.seed()

    #     # get photon budget and photobleaching-time
    #     tau = rng.choice(tau_bleach, N_part, p=p_bleach)

    #     for i in range(N_part):

    #         # get photon budget and photobleaching-time
    #         photons = (tau[i]/tau0)*photon0

    #         N = 10000 #int((end - start)/dt)

    #         time  = numpy.array([j*dt + start for j in range(N)])
    #         state = numpy.zeros(shape=(N))

    #         t0 = start
    #         t1 = t0 + tau[i]

    #         N0 = (numpy.abs(time - t0)).argmin()
    #         N1 = (numpy.abs(time - t1)).argmin()
    #         state[N0:N1] = numpy.ones(shape=(N1-N0))

    #         if (self.photoblinking_switch == True):

    #             # ON-state
    #             t_on  = rng.choice(tau_on,  100, p=p_on)
    #             # OFF-state
    #             t_off = rng.choice(tau_off, 100, p=p_off)

    #             k = (numpy.abs(numpy.cumsum(t_on) - tau[i])).argmin()

    #             # merge t_on/off arrays
    #             t_blink = numpy.array([t_on[0:k], t_off[0:k]])
    #             t = numpy.reshape(t_blink, 2*k, order='F')

    #             i_on  = N0

    #             for j in range(k):

    #                 f_on  = i_on  + int(t[j]/dt)
    #                 i_off = f_on
    #                 f_off = i_off + int(t[j+1]/dt)

    #                 state[i_on:f_on]   = 1
    #                 state[i_off:f_off] = 0

    #                 i_on = f_off

    #         # set photon budget and fluorescence state
    #         budget.append(photons)
    #         state_act.append(state)

    #     self.fluorescence_bleach = numpy.array(tau)
    #     self.fluorescence_budget = numpy.array(budget)
    #     self.fluorescence_state  = numpy.array(state_act)

class EPIFMConfigs:

    def __init__(self, config, rng=None):
        self.initialize(config, rng=rng)

    def initialize(self, config, rng=None):
        """Initialize based on the given config.

        Args:
            config (Configuration): A config object.
            rng (numpy.RandomState, optional): A random number generator
                for initializing this class. Defaults to `None`.

        """
        if config.type.lower() != 'epifm':
            raise ValueError("An invalid type [{}] was given. 'epifm' is required.".format(config.type))

        if rng is None:
            warnings.warn('A random number generator [rng] is not given.')
            rng = numpy.random.RandomState()

        self.set_fluorophore(**config.fluorophore)  # => self.__wave_length

        self.set_shutter(**config.shutter)
        self.set_light_source(**config.light_source)
        self.set_dichroic_mirror(**config.dichroic_mirror)

        self.image_magnification = config.magnification
        _log.info('--- Magnification: x {}'.format(self.image_magnification))

        self.set_detector(**config.detector)
        self.set_analog_to_digital_converter(rng=rng, **config.analog_to_digital_converter)
        self.set_excitation_filter(**config.excitation_filter)
        self.set_emission_filter(**config.emission_filter)

        # self.fluorophore_psf = self.get_PSF_detector()
        self.fluorophore_psf = PointSpreadingFunction(
            config.fluorophore.radial_cutoff, self.psf_radial_width, config.fluorophore.depth_cutoff,
            self.fluorophore_type, self.psf_wavelength)

        self.effects = PhysicalEffectConfigs(config.effects)

    def set_shutter(self, start_time=None, end_time=None, time_open=None, time_lapse=None, switch=True):
        self.shutter_switch = switch
        self.shutter_start_time = start_time
        self.shutter_end_time = end_time

        _log.info('--- Shutter:')
        _log.info('    Start-Time = {} sec'.format(self.shutter_start_time))
        _log.info('    End-Time   = {} sec'.format(self.shutter_end_time))

    def set_light_source(self, type=None, wave_length=None, flux_density=None, radius=None, angle=None, switch=True):
        self.source_switch = switch
        self.source_type = type
        self.source_wavelength = wave_length
        self.source_flux_density = flux_density
        self.source_radius = radius
        self.source_angle = angle

        _log.info('--- Light Source:{}'.format(self.source_type))
        _log.info('    Wave Length = {} m'.format(self.source_wavelength))
        _log.info('    Beam Flux Density = {} W/m2'.format(self.source_flux_density))
        _log.info('    1/e2 Radius = {} m'.format(self.source_radius))
        _log.info('    Angle = {} radian'.format(self.source_angle))

    def set_fluorophore(
            self, type=None, wave_length=None, normalization=None, radius=None, radial_width=None,
            min_wave_length=None, max_wave_length=None, radial_cutoff=None, depth_cutoff=None):
        self.__wave_length = numpy.arange(min_wave_length, max_wave_length, 1e-9, dtype=float)

        self.fluorophore_type = type
        self.fluorophore_radius = radius
        self.psf_normalization = normalization

        if type == 'Gaussian':
            N = len(self.__wave_length)
            fluorophore_excitation = numpy.zeros(N, dtype=float)
            fluorophore_emission = numpy.zeros(N, dtype=float)
            idx = (numpy.abs(self.__wave_length - wave_length)).argmin()
            index_ex = index_em = idx
            self.psf_radial_width = radial_width
        else:
            (fluorophore_excitation, fluorophore_emission) = io.read_fluorophore_catalog(type)
            fluorophore_excitation = self.calculate_efficiency(fluorophore_excitation, self.__wave_length)
            fluorophore_emission = self.calculate_efficiency(fluorophore_emission, self.__wave_length)
            fluorophore_excitation = numpy.array(fluorophore_excitation)
            fluorophore_emission = numpy.array(fluorophore_emission)
            index_ex = fluorophore_excitation.argmax()
            index_em = fluorophore_emission.argmax()
            if wave_length is not None:
                warnings.warn('The given wave length [{}] was ignored'.format(wave_length))
            self.psf_radial_width = None  #XXX: This is only required for Gaussian

        fluorophore_excitation[index_ex] = 100
        fluorophore_emission[index_em] = 100
        self.fluoex_eff = fluorophore_excitation
        self.fluoem_eff = fluorophore_emission
        fluorophore_excitation /= sum(fluorophore_excitation)
        fluorophore_emission /= sum(fluorophore_emission)
        self.fluoex_norm = fluorophore_excitation
        self.fluoem_norm = fluorophore_emission
        self.psf_wavelength = self.__wave_length[index_em]

        _log.info('--- Fluorophore: {} PSF'.format(self.fluorophore_type))
        _log.info('    Wave Length   =  {} m'.format(self.psf_wavelength))
        _log.info('    Normalization =  {}'.format(self.psf_normalization))
        _log.info('    Fluorophore radius =  {} m'.format(self.fluorophore_radius))
        if self.psf_radial_width is not None:
            _log.info('    Lateral Width =  {} m'.format(self.psf_radial_width))
            # _log.info('    Lateral Width =  {} m'.format(self.psf_width[0]))
            # _log.info('    Axial Width =  {} m'.format(self.psf_width[1]))
        _log.info('    PSF Normalization Factor =  {}'.format(self.psf_normalization))
        _log.info('    Emission  : Wave Length =  {} m'.format(self.psf_wavelength))

    def set_dichroic_mirror(self, type=None, switch=True):
        self.dichroic_switch = switch
        if type is not None:
            dichroic_mirror = io.read_dichroic_catalog(type)
            self.dichroic_eff = self.calculate_efficiency(dichroic_mirror, self.__wave_length)
        else:
            self.dichroic_eff = numpy.zeros(len(self.__wave_length), dtype=float)
        _log.info('--- Dichroic Mirror:')

    def set_detector(
            self, type=None, image_size=None, pixel_length=None, exposure_time=None, focal_point=None,
            QE=None, readout_noise=None, dark_count=None, emgain=None, switch=True):
        self.detector_switch = switch
        self.detector_type = type
        self.detector_image_size = image_size
        self.detector_pixel_length = pixel_length
        self.detector_exposure_time = exposure_time
        self.detector_focal_point = focal_point
        self.detector_qeff = QE
        self.detector_readout_noise = readout_noise
        self.detector_dark_count = dark_count
        self.detector_emgain = emgain
        # self.detector_base_position = base_position
        self.detector_focal_point = focal_point

        _log.info('--- Detector:  {}'.format(self.detector_type))
        if self.detector_image_size is not None:
            _log.info('    Image Size  =  {} x {}'.format(self.detector_image_size[0], self.detector_image_size[1]))
        _log.info('    Pixel Size  =  {} m/pixel'.format(self.detector_pixel_length))
        _log.info('    Focal Point =  {}'.format(self.detector_focal_point))
        _log.info('    Exposure Time =  {} sec'.format(self.detector_exposure_time))
        if self.detector_qeff is not None:
            _log.info('    Quantum Efficiency =  {} %'.format(100 * self.detector_qeff))
        _log.info('    Readout Noise =  {} electron'.format(self.detector_readout_noise))
        _log.info('    Dark Count =  {} electron/sec'.format(self.detector_dark_count))
        _log.info('    EM gain = x {}'.format(self.detector_emgain))
        # _log.info('    Position    =  {}'.format(self.detector_base_position))
        _log.info('Focal Center: {}'.format(self.detector_focal_point))

    def set_analog_to_digital_converter(self, *, rng, bit=None, offset=None, fullwell=None, type=None, count=None):
        self.ADConverter_bit = bit
        self.ADConverter_fullwell = fullwell
        self.ADConverter_fpn_type = type
        self.ADConverter_fpn_count = count

        offset, gain = self.calculate_analog_to_digital_converter_gain(offset, rng)
        self.ADConverter_offset = offset
        self.ADConverter_gain = gain

        _log.info('--- A/D Converter: %d-bit' % (self.ADConverter_bit))
        # _log.info('    Gain = %.3f electron/count' % (self.ADConverter_gain))
        # _log.info('    Offset =  {} count'.format(self.ADConverter_offset))
        _log.info('    Fullwell =  {} electron'.format(self.ADConverter_fullwell))
        _log.info('    {}-Fixed Pattern Noise: {} count'.format(self.ADConverter_fpn_type, self.ADConverter_fpn_count))

    def calculate_analog_to_digital_converter_gain(self, ADC0, rng):
        # image pixel-size
        Nw_pixel = self.detector_image_size[0]
        Nh_pixel = self.detector_image_size[1]

        # set ADC parameters
        bit  = self.ADConverter_bit
        fullwell = self.ADConverter_fullwell
        # ADC0 = self.ADConverter_offset

        # set Fixed-Pattern noise
        FPN_type = self.ADConverter_fpn_type
        FPN_count = self.ADConverter_fpn_count

        # if FPN_type is None:
        if FPN_type == 'none':
            offset = numpy.full(Nw_pixel * Nh_pixel, ADC0)
        elif FPN_type == 'pixel':
            if rng is None:
                raise RuntimeError('A random number generator is required.')
            offset = numpy.rint(rng.normal(ADC0, FPN_count, Nw_pixel * Nh_pixel))
        elif FPN_type == 'column':
            if rng is None:
                raise RuntimeError('A random number generator is required.')
            column = rng.normal(ADC0, FPN_count, Nh_pixel)
            temporal = numpy.tile(column, (1, Nw_pixel))
            offset = numpy.rint(temporal.reshape(Nh_pixel * Nw_pixel))
        else:
            raise ValueError("FPN type [{}] is invalid ['pixel', 'column' or 'none']".format(FPN_type))

        # set ADC gain
        # gain = numpy.array(map(lambda x: (fullwell - 0.0) / (pow(2.0, bit) - x), offset))
        gain = (fullwell - 0.0) / (pow(2.0, bit) - offset)

        offset = offset.reshape([Nw_pixel, Nh_pixel])
        gain = gain.reshape([Nw_pixel, Nh_pixel])

        return (offset, gain)

    def set_excitation_filter(self, type=None, switch=True):
        self.excitation_switch = switch
        if type is not None:
            excitation_filter = io.read_excitation_catalog(type)
            self.excitation_eff = self.calculate_efficiency(excitation_filter, self.__wave_length)
        else:
            self.excitation_eff = numpy.zeros(len(self.__wave_length), dtype=float)
        _log.info('--- Excitation Filter:')

    def set_emission_filter(self, type=None, switch=True):
        self.emission_switch = switch
        if type is not None:
            emission_filter = io.read_emission_catalog(type)
            self.emission_eff = self.calculate_efficiency(emission_filter, self.__wave_length)
        else:
            self.emission_eff = numpy.zeros(len(self.__wave_length), dtype=float)
        _log.info('--- Emission Filter:')

    @staticmethod
    def calculate_efficiency(data, wave_length):
        data = numpy.array(data, dtype = 'float')
        data = data[data[: , 0] % 1 == 0, :]

        efficiency = numpy.zeros(len(wave_length))

        wave_length = numpy.round(wave_length / 1e-9).astype(int)  #XXX: numpy.array(dtype=int)
        idx1 = numpy.in1d(wave_length, data[:, 0])
        idx2 = numpy.in1d(numpy.array(data[:, 0]), wave_length)

        efficiency[idx1] = data[idx2, 1]

        return efficiency.tolist()

class _EPIFMSimulator:
    '''
    A class of the simulator for Epifluorescence microscopy (EPI).
    '''

    def __init__(self, configs, environ=None):
        """A constructor of _EPIFMSimulator.

        Args:
            config (Config, optional): A config object for initializing this class.
            rng (numpy.RandomState, optional): A random number generator
                for initializing this class.

        """
        self.configs = configs

        if environ is not None:
            self.environ = environ

    def generate_frames(
            self, input_data, num_frames, start_time=0.0, exposure_time=None,
            rng=None, processes=None):
        """Output all images from the given particle data.

        Args:
            input_data (list): An input data. A list of pairs of time and a list of particles.
                Each particle is represented as a list of numbers: a coordinate (a triplet of floats),
                molecule id, and a state of fluorecent.
                The number of particles in each frame must be static.
            num_frames (int): The number of frames taken.
            start_time (float, optional): A time to start detecting.
                Defaults to 0.0.
            exposure_time (float, optional): An exposure time.
                Defaults to `detector_exposure_time` in the configuration.
            rng (numpy.RandomState, optional): A random number generator.

        """
        if rng is None:
            _log.info('A random number generator was initialized.')
            rng = numpy.random.RandomState()

        fluorescence_states = None
        if self.configs.effects.photobleaching_switch:
            fluorescence_states = {}

        exposure_time = exposure_time or self.configs.detector_exposure_time

        for frame_index in range(num_frames):
            camera, infodict = self.output_frame(
                    input_data, frame_index=frame_index, start_time=start_time, exposure_time=exposure_time,
                    fluorescence_states=fluorescence_states, rng=rng, processes=processes)
            yield (camera, infodict)

    def output_frames(
            self, input_data, num_frames, start_time=0.0, exposure_time=None,
            pathto='./images', data_fmt='image_%07d.npy', true_fmt='true_%07d.npy',
            image_fmt='image_%07d.png', cmin=None, cmax=None, low=None, high=None,
            rng=None, processes=None):
        """Output all images from the given particle data.

        Args:
            input_data (list): An input data. A list of pairs of time and a list of particles.
                Each particle is represented as a list of numbers: a coordinate (a triplet of floats),
                molecule id, and a state of fluorecent.
                The number of particles in each frame must be static.
            num_frames (int): The number of frames taken.
            start_time (float, optional): A time to start detecting.
                Defaults to 0.0.
            exposure_time (float, optional): An exposure time.
                Defaults to `detector_exposure_time` in the configuration.
            pathto (str, optional): A path to save images and ndarrays. Defaults to './images'.
            image_fmt (str, optional): A format of the filename to save 8-bit images.
                An int is available as its frame index. Defaults to 'image_%07d.png'.
                If None, no file is saved.
            data_fmt (str, optional): A format of the filename to save the detected photon counts.
                An int is available as its frame index. Defaults to 'image_%07d.npy'.
                If None, no file is saved.
            true_fmt (str, optional): A format of the filename to save true data.
                An int is available as its frame index. Defaults to 'true_%07d.npy'.
                If None, no file is saved.
            cmin (int, optional): A minimal value used to generate 8-bit images.
            cmax (int, optional): A maximum value used to generate 8-bit images.
            low (int, optional): A minimal value used to generate 8-bit images.
                Defaults to 0.
            high (int, optional): A maximum value used to generate 8-bit images.
                Defaults to 255.
            rng (numpy.RandomState, optional): A random number generator.

        """
        # Check and create the folders for image and output files.
        if not os.path.exists(pathto):
            os.makedirs(pathto)

        low = 0 if low is None else low
        high = 255 if high is None else high

        if rng is None:
            _log.info('A random number generator was initialized.')
            rng = numpy.random.RandomState()

        results = []
        for camera, infodict in self.generate_frames(
                input_data, frame_index=frame_index, start_time=start_time, exposure_time=exposure_time,
                fluorescence_states=fluorescence_states, rng=rng, processes=processes):
            # save photon counts to numpy-binary file
            if data_fmt is not None:
                data_file_name = os.path.join(pathto, data_fmt % (frame_index))
                numpy.save(data_file_name, camera)

            # save true-dataset to numpy-binary file
            if true_fmt is not None:
                raise RuntimeError("Not supported.")
                # true_file_name = os.path.join(pathto, true_fmt % (frame_index))
                # numpy.save(true_file_name, true_data)

            # save images to numpy-binary file
            if image_fmt is not None:
                image_file_name = os.path.join(pathto, image_fmt % (frame_index))
                Image(camera[: , : , 1]).as_8bit(cmin, cmax, low, high).save(image_file_name)

            results.append((camera, infodict))
        return results

    def output_frame(
            self, input_data, frame_index=0, start_time=0.0, exposure_time=None,
            fluorescence_states=None, rng=None, processes=None):
        """Output an image from the given particle data.

        Args:
            input_data (list): An input data. A list of pairs of time and a list of particles.
                Each particle is represented as a list of numbers: a coordinate (x, y, z),
                molecule id, and a state of fluorecent.
                The number of particles in each frame must be static.
            frame_index (int, optional): An index of the frame. The value must be 0 and more.
                Defaults to 0.
            start_time (float, optional): A time to start detecting.
                Defaults to 0.0.
            exposure_time (float, optional): An exposure time.
                Defaults to `detector_exposure_time` in the configuration.
            rng (numpy.RandomState, optional): A random number generator.

        """
        processes = processes or (1 if self.environ is None else self.environ.processes)
        exposure_time = exposure_time or self.configs.detector_exposure_time

        _log.info('processes = {}.'.format(processes))

        if rng is None:
            _log.info('A random number generator was initialized.')
            rng = numpy.random.RandomState()

        times = numpy.array([t for t, _ in input_data])
        t = start_time + exposure_time * frame_index

        if self.configs.shutter_switch:
            t = max(t, self.configs.shutter_start_time)
            exposure_time = max(0.0, min(t + exposure_time, self.configs.shutter_end_time))

        start_index = numpy.searchsorted(times, t, side='right')
        if start_index != 0:
            start_index -= 1
        stop_index = numpy.searchsorted(times, t + exposure_time, side='left')
        if times[start_index] > t:
            warnings.warn("No data input for interval [{}, {}]".format(t, times[start_index]))

        frame_data = input_data[start_index: stop_index]

        _log.info('time: {} sec ({})'.format(t, frame_index))

        # focal point
        p_0 = numpy.asarray(self.configs.detector_focal_point)  # meter-scale
        _log.info('center (width, height): {} {}'.format(p_0[0], p_0[1]))

        # beam position: Assuming beam position = focal point (for temporary)
        p_b = copy.copy(p_0)

        # camera pixels
        Nw_pixel, Nh_pixel = self.configs.detector_image_size  # pixels
        pixel_length = self.configs.detector_pixel_length / self.configs.image_magnification  # m-scale
        camera_pixel = numpy.zeros((Nw_pixel, Nh_pixel, 2))
        _log.info('scaling [m/pixel]: {}'.format(pixel_length))

        optinfo = {}

        # loop for frame data
        for i, (i_time, particles) in enumerate(frame_data):
            # When the first i_time is less than the start time given, exposure_time is shorter than expected
            current_time = i_time if i != 0 else t
            next_time = frame_data[i + 1][0] if i + 1 < len(frame_data) else t + exposure_time
            unit_time = next_time - current_time

            _log.info('    {:03d}-th file in {:03d}-th frame: {} + {} sec'.format(
                i, frame_index, current_time, unit_time))

            if unit_time < 1e-13:  #XXX: Skip merging when the exposure time is too short
                continue

            # loop for particles
            expected, optinfo_, states = self.get_molecule_plane(
                    particles, shape=camera_pixel.shape[: 2], p_b=p_b, p_0=p_0, unit_time=unit_time,
                    optional_info=optinfo, fluorescence_states=fluorescence_states, rng=rng, processes=processes)
            camera_pixel[:, :, 0] += expected
            optinfo.update(optinfo_)
            if fluorescence_states is not None:
                fluorescence_states.update(states)

        _log.info('Total camera_pixel = {:3e}'.format(camera_pixel[:, :, 0].sum()))
        _log.info('{} optoinfos are registered.'.format(len(optinfo)))

        for m_id in optinfo:
            optinfo[m_id][1] /= exposure_time  # photon state
            optinfo[m_id][2: 6] /= optinfo[m_id][0]  # Calculating time averages

            # Transforming CoM (X and Y) from coordinates to pixels.
            # The coordinate is positioned at the center of a pixel, rather than the corner.
            # Therefore, shifting not `Nw_pixel * 0.5`, but `(Nw_pixel - 1) * 0.5`.
            # See also `PointSpreadingFunction.overlay_signal`_
            optinfo[m_id][2] = (optinfo[m_id][2] - p_0[1]) / pixel_length + (Nw_pixel - 1) * 0.5
            optinfo[m_id][3] = (optinfo[m_id][3] - p_0[2]) / pixel_length + (Nh_pixel - 1) * 0.5

        # apply detector effects
        _log.info("Apply detector effects.")
        camera = self.__detector_output(rng, camera_pixel, processes=processes)
        # return (camera, optinfo)
        infodict = dict(true_data=optinfo)
        if fluorescence_states is not None:
            infodict['fluorescence_states'] = copy.copy(fluorescence_states)
        return (camera, infodict)

    def get_molecule_plane(
            self, particles, shape, p_b, p_0, unit_time,
            optional_info, fluorescence_states, rng=None, processes=None):
        expected = numpy.zeros(shape)
        optinfo = {int(m_id): optional_info[int(m_id)]
                for _, _, _, m_id, _ in particles if int(m_id) in optional_info}
        if fluorescence_states is None:
            states = None
        else:
            states = {int(m_id): fluorescence_states[int(m_id)]
                    for _, _, _, m_id, _ in particles if int(m_id) in fluorescence_states}

        if processes is not None and processes > 1:
            if fluorescence_states is not None and self.configs.effects.photobleaching_switch:
                amplitude0, _ = self.snells_law()
                N_emit0 = self.__get_emit_photons(amplitude0, 1.0)
                half_life = self.configs.effects.photobleaching_half_life
                for p in particles:
                    m_id = int(p[3])
                    if m_id not in states:
                        states[m_id] = self.get_photon_budget(
                                N_emit0, half_life, rng=rng)

            func = functools.partial(
                    self.get_molecule_plane,
                    shape=shape, p_b=p_b, p_0=p_0, unit_time=unit_time,
                    optional_info=optinfo, fluorescence_states=states, rng=None, processes=None)
            with Pool(processes) as pool:
                results = pool.map(func, numpy.array_split(particles, processes))
            for expected_, optinfo_, states_ in results:
                expected += expected_
                optinfo.update(optinfo_)
                if states is not None:
                    states.update(states_)
        else:
            for i in range(len(particles)):
                self.__overlay_molecule_plane(
                        expected, particles[i], p_b, p_0, unit_time, optinfo, states, rng)
        return expected, optinfo, states

    def __overlay_molecule_plane(
            self, expected, particle_i, _p_b, p_0, unit_time,
            optional_info, fluorescence_states=None, rng=None):
        # m-scale
        # p_b (ndarray): beam position (assumed to be the same with focal center), but not used.
        # p_0 (ndarray): focal center

        # particles coordinate, molecule id, photon state
        x, y, z, m_id, p_state = particle_i
        m_id = int(m_id)

        p_i = numpy.array([x, y, z])

        # Snell's law
        amplitude, penet_depth = self.snells_law()  #XXX: This might depends on p_b and p_i

        # particles coordinate in real(nm) scale
        _, radial, depth = cylindrical_coordinate(p_i, p_0)

        # get exponential amplitude (only for TIRFM-configuration)
        # depth must be positve in case of TIRFM
        #XXX: amplitude = amplitude * numpy.exp(-depth / penet_depth)
        amplitude = amplitude * numpy.exp(-abs(depth) / penet_depth)

        # get the number of photons emitted
        N_emit = self.__get_emit_photons(amplitude, unit_time)

        if fluorescence_states is not None:
            if self.configs.effects.photobleaching_switch:
                if m_id not in fluorescence_states:
                    assert rng is not None
                    amplitude0, _ = self.snells_law()
                    N_emit0 = self.__get_emit_photons(amplitude0, 1.0)
                    fluorescence_states[m_id] = self.get_photon_budget(
                            N_emit0, self.configs.effects.photobleaching_half_life, rng=rng)  #XXX: rng is only required here
                budget = fluorescence_states[m_id] - N_emit
                if budget <= 0:
                    budget = 0
                    p_state = 0.0  #XXX
                fluorescence_states[m_id] = budget

        # Photon Transmission Efficiency
        normalization = self.configs.fluoem_norm
        if self.configs.dichroic_switch:
            normalization *= 0.01 * self.configs.dichroic_eff
        if self.configs.emission_switch:
            normalization *= 0.01 * self.configs.emission_eff
        normalization = numpy.sum(normalization) * self.configs.psf_normalization
        normalization *= p_state * N_emit / (4.0 * numpy.pi)

        # add signal matrix to image plane
        pixel_length = self.configs.detector_pixel_length / self.configs.image_magnification
        self.configs.fluorophore_psf.overlay_signal(expected, p_i - p_0, pixel_length, normalization)

        # set true-dataset
        if m_id not in optional_info:
            optional_info[m_id] = numpy.zeros(8, dtype=numpy.float64)
        optional_info[m_id] += numpy.array([
            unit_time,
            unit_time * p_state,  # photon-state
            unit_time * p_i[1],  # X-coordinate in the image-plane
            unit_time * p_i[2],  # Y-coordinate in the image-plane
            unit_time * p_i[1],  # X-coordinate in the image-plane
            unit_time * p_i[2],  # Y-coordinate in the image-plane
            unit_time * depth,  # Depth from focal-plane
            normalization,  # normalization
            ])

    def __get_emit_photons(self, amplitude, unit_time):
        # Absorption coeff [1/(cm M)]
        abs_coeff = self.configs.effects.abs_coefficient
        # Quantum yield
        QY = self.configs.effects.quantum_yield
        fluorophore_radius = self.configs.fluorophore_radius
        return self.get_emit_photons(amplitude, unit_time, abs_coeff, QY, fluorophore_radius)

    @staticmethod
    def get_emit_photons(amplitude, unit_time, abs_coeff, QY, fluorophore_radius):
        # Cross-section [m2]
        x_sec = numpy.log(10) * abs_coeff * 0.1 / constants.N_A

        # get the number of absorption photons: [#/(m2 sec)]*[m2]*[sec]
        n_abs = amplitude * x_sec * unit_time

        # Beer-Lamberts law: A = log(I0 / I) = abs coef. * concentration * path length ([m2]*[#/m3]*[m])
        fluorophore_volume = (4.0 / 3.0) * numpy.pi * numpy.power(fluorophore_radius, 3)
        fluorophore_depth  = 2.0 * fluorophore_radius

        A = (abs_coeff * 0.1 / constants.N_A) * (1.0 / fluorophore_volume) * fluorophore_depth

        # get the number of photons emitted
        n_emit = QY * n_abs * (1.0 - numpy.power(10.0, -A))

        return n_emit

    def snells_law(self):
        """Snell's law.

        Returns:
            float: Amplitude
            float: Penetration depth

        """
        # x_0, y_0, z_0 = p_0
        # x_i, y_i, z_i = p_i

        # Illumination: Assume that uniform illumination (No gaussian)
        # flux density, [W / m ** 2] = [J / s / m ** 2] = [kg / s ** 3]
        P_0 = self.configs.source_flux_density

        # single photon energy
        wave_length = self.configs.source_wavelength  # m
        E_wl = constants.hc / wave_length

        # photon flux density [photons/sec/m2]
        N_0 = P_0 / E_wl

        # Incident beam: Amplitude
        A2_Is = N_0
        A2_Ip = N_0

        # incident beam angle (radian)
        angle = self.configs.source_angle
        # theta_in = (angle / 180.) * numpy.pi
        theta_in = angle

        sin_th1 = numpy.sin(theta_in)
        cos_th1 = numpy.cos(theta_in)

        sin = sin_th1
        cos = cos_th1
        sin2 = sin ** 2
        cos2 = cos ** 2

        # index of refraction
        n_1 = 1.46  # fused silica
        n_2 = 1.384 # cell
        # n_3 = 1.337 # culture medium

        r  = n_2 / n_1
        r2 = r ** 2

        # Epi-illumination at apical surface
        if sin2 / r2 < 1:
            amplitude = N_0  # = (A2_Ip + A2_Is) / 2
            penetration_depth = numpy.inf
        else:
            # TIRF-illumination at basal cell-surface
            # Evanescent field: Amplitude and Depth
            # Assume that the s-polar direction is parallel to y-axis
            A2_x = A2_Ip * (4 * cos2 * (sin2 - r2) / (r2 ** 2 *cos2 + sin2 - r2))
            A2_y = A2_Is * (4 * cos2 / (1 - r2))
            A2_z = A2_Ip * (4 * cos2 * sin2 / (r2 ** 2 * cos2 + sin2 - r2))

            # set illumination amplitude
            A2_Tp = A2_x + A2_z
            A2_Ts = A2_y
            amplitude = (A2_Tp + A2_Ts) / 2

            penetration_depth = wave_length / (4.0 * numpy.pi * numpy.sqrt(n_1 ** 2 * sin2 - n_2 ** 2))

        return amplitude, penetration_depth

    def __detector_output(self, rng, camera_pixel, processes=1):
        ## Detector: Quantum Efficiency
        # index = int(self.configs.psf_wavelength / 1e-9) - int(self.configs.wave_length[0] / 1e-9)
        # QE = self.configs.detector_qeff[index]
        QE = self.configs.detector_qeff
        ## get signal (photons)
        photons = camera_pixel[:, :, 0]
        ## get constant background (photoelectrons)
        if self.configs.effects.background_switch:
            photons += self.configs.effects.background_mean
        ## get signal (expectation)
        expected = QE * photons

        ## conversion: photon --> photoelectron --> ADC count
        ## select Camera type
        if self.configs.detector_type == "CMOS":
            noise = CMOS.get_noise(expected.shape, rng=rng)
            signal = CMOS.get_signal(expected, rng=rng, processes=processes)
        elif self.configs.detector_type == "EMCCD":
            noise = EMCCD.get_noise(expected.shape, readout_noise=self.configs.detector_readout_noise, rng=rng)
            signal = EMCCD.get_signal(
                    expected,
                    emgain=self.configs.detector_emgain,
                    rng=rng,
                    processes=processes)
        elif self.configs.detector_type == "CCD":
            noise = CCD.get_noise(expected.shape, readout_noise=self.configs.detector_readout_noise, rng=rng)
            signal = CCD.get_signal(expected, rng=rng, processes=processes)
        else:
            raise RuntimeError(
                    "Unknown detector type was given [{}]. ".format(self.configs.detector_type)
                    + "Use either one of 'CMOS', 'CCD' or 'EMCCD'.")

        camera_pixel[:, :, 0] = expected
        camera_pixel[:, :, 1] = self.__get_analog_to_digital_converter_counts(
                signal + noise,
                fullwell=self.configs.ADConverter_fullwell,
                gain=self.configs.ADConverter_gain,
                offset=self.configs.ADConverter_offset,
                bit=self.configs.ADConverter_bit)
        return camera_pixel

    @staticmethod
    def __get_analog_to_digital_converter_counts(photoelectron, fullwell, gain, offset, bit):
        # check non-linearity
        ADC = photoelectron.copy()
        ADC[ADC > fullwell] = fullwell

        # convert photoelectron to ADC counts
        ADC_max = 2 ** bit - 1
        ADC /= gain
        ADC += offset
        ADC[ADC > ADC_max] = ADC_max
        ADC[ADC < 0] = 0
        return ADC

    @staticmethod
    def get_photon_budget(N_emit0, half_life, rng):
        beta = half_life / numpy.log(2.0)
        bleaching_time = rng.exponential(scale=beta, size=None)
        budget = bleaching_time * N_emit0
        return budget



import scopyon
import numpy.random


def main():
    config = scopyon.DefaultConfiguration()
    pixel_length = config.default.detector.pixel_length / config.default.magnification
    L_2 = config.default.detector.image_size[0] * pixel_length * 0.5

    rng = numpy.random.RandomState(123)
    N = 250
    inputs = rng.uniform(-L_2, +L_2, size=(N, 2))

    img1 = scopyon.form_image(inputs[: 200], config=config, rng=rng)
    img2 = scopyon.form_image(inputs[150: ], config=config, rng=rng)
    img = scopyon.Image.RGB(red=img1, green=img2)
    img.save("twocolor_000.png")


if __name__ == "__main__":
    import logging
    # logging.basicConfig(level=logging.DEBUG)
    handler = logging.StreamHandler()
    fmt = logging.Formatter(logging.BASIC_FORMAT)
    handler.setFormatter(fmt)
    log = logging.getLogger('scopyon')
    log.addHandler(handler)
    log.setLevel(logging.DEBUG)

    main()



import unittest

# from bioimaging import epifm


class TestTirfm(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test1(self):
        pass


if __name__ == '__main__':
    unittest.main()



import collections.abc
import numbers
import warnings

import numpy

import scopyon._epifm
import scopyon.config
import scopyon.image

from logging import getLogger
_log = getLogger(__name__)

__all__ = [
    "EnvironSettings", "EPIFMSimulator",
    "form_image", "generate_images", "create_simulator"
    ]


class EnvironSettings:

    def __init__(self, config):
        self.initialize(config)

    def initialize(self, config):
        self.processes = config.processes

class EPIFMSimulator(object):

    def __init__(self, config=None, method=None, rng=None):
        """ Constructor

        Args:
            config (Configuration or str, optional): Configurations.
                The default is None.
            method (str, optional): A name of method used.
                The default is None (config.default).
            rng (numpy.RandomState, optional): A random number generator.
                The default is None.
        """
        if config is None:
            config = scopyon.config.Configuration()
        elif isinstance(config, str):
            config = scopyon.config.Configuration(filename=config)
        elif not isinstance(config, scopyon.config.Configuration):
            raise TypeError("Configuration or str must be given [{}].".format(type(config)))

        if rng is None:
            warnings.warn('A random number generator is not given.')
            rng = numpy.random.RandomState()

        self.__config = config
        self.__method = method or config.default.lower()
        self.__rng = rng

    def base(self):
        return scopyon._epifm._EPIFMSimulator(
                configs=scopyon._epifm.EPIFMConfigs(self.__config[self.__method], rng=self.__rng),
                environ=EnvironSettings(self.__config.environ))

    def __format_data(self, inputs):
        assert isinstance(inputs, numpy.ndarray)

        if inputs.ndim != 2:
            raise ValueError("The given 'inputs' has wrong dimension.")

        if inputs.shape[1] == 2:
            data = numpy.hstack((
                numpy.zeros((inputs.shape[0], 1)),
                inputs * self.__config.preprocessing.scale))
            data = numpy.hstack((
                data,
                numpy.zeros((data.shape[0], 2))))
            data[:, 3] = numpy.arange(inputs.shape[0])  #FIXME: Molecule ID
            data[:, 4] = 1.0  # Photon state
        elif inputs.shape[1] == 3:
            origin = numpy.array(self.__config.preprocessing.origin)
            data = inputs * self.__config.preprocessing.scale - origin
            unit_z = numpy.cross(
                self.__config.preprocessing.unit_x,
                self.__config.preprocessing.unit_y)
            data = numpy.hstack((
                numpy.dot(data, unit_z).reshape((-1, 1)),
                numpy.dot(data, self.__config.preprocessing.unit_x).reshape((-1, 1)),
                numpy.dot(data, self.__config.preprocessing.unit_y).reshape((-1, 1)),
                ))
            data = numpy.hstack((
                data,
                numpy.zeros((data.shape[0], 2))))
            data[:, 3] = numpy.arange(inputs.shape[0])  #FIXME: Molecule ID
            data[:, 4] = 1.0  # Photon state
        elif inputs.shape[1] == 4:
            data = numpy.hstack((
                numpy.zeros((inputs.shape[0], 1)),
                inputs[:, : 2] * self.__config.preprocessing.scale,
                inputs[:, 2: ]))
        elif inputs.shape[1] == 5:
            origin = numpy.array(self.__config.preprocessing.origin)
            data = inputs[:, : 3] * self.__config.preprocessing.scale - origin
            unit_z = numpy.cross(
                self.__config.preprocessing.unit_x,
                self.__config.preprocessing.unit_y)
            data = numpy.hstack((
                numpy.dot(data, unit_z).reshape((-1, 1)),
                numpy.dot(data, self.__config.preprocessing.unit_x).reshape((-1, 1)),
                numpy.dot(data, self.__config.preprocessing.unit_y).reshape((-1, 1)),
                inputs[:, 3: ]))
        else:
            raise ValueError("The given 'inputs' has wrong shape.")
        return data

    def __format_inputs(self, inputs):
        if isinstance(inputs, numpy.ndarray):
            # Check this first. numpy.ndarray is even Iterable.
            data = ((0.0, self.__format_data(inputs)), )
        elif isinstance(inputs, collections.abc.Iterable):
            data = []
            for elem in inputs:
                if not (isinstance(elem, (tuple, list)) and len(elem) == 2
                        and isinstance(elem[0], numbers.Real) and isinstance(elem[1], numpy.ndarray)):
                    raise ValueError("The given 'inputs' has wrong type.")
                data.append((elem[0], self.__format_data(elem[1])))
        else:
            raise TypeError(
                    "Invalid argument was given [{}]."
                    " A ndarray is expected.".format(type(inputs)))
        return data

    def form_image(
            self, inputs, start_time=0.0, exposure_time=None,
            full_output=False):
        """Form image.

        Args:
            inputs (array_like): A list of points. The shape must be '(n, 3)', where
                'n' means the number of points.
            start_time (float, optional): A time to start detecting.
                Defaults to 0.0.
            exposure_time (float, optional): An exposure time.
                Defaults to `detector.exposure_time` in the configuration.
            full_output (bool, optional):
                True to return a dictionary containing optional outputs.

        Returns:
            Image: An image.
            dict: only returned if full_output == True
                A dictionary containing additional information.
                'expectation' is 2-dimensional ndarray.
                'true_data' is a dictionary containing a pair of a molecule ID and ndarray.
                The array has 7 elements, which are exposure time, photon state, X and Y in pixels,
                X and Y in meter-scale, and normalization.
        """
        data = self.__format_inputs(inputs)
        base = self.base()
        camera, infodict = base.output_frame(
                data, start_time=start_time, exposure_time=exposure_time, rng=self.__rng)
        # camera[:, :, 0] => expected
        # camera[:, :, 1] => ADC
        img = scopyon.image.Image(camera[:, :, 1])
        if full_output:
            infodict.update(dict(expectation=camera[:, :, 0]))
            return img, infodict
        return img

    def generate_images(
            self, inputs, num_frames, start_time=0.0, exposure_time=None,
            full_output=False):
        """Form image.

        Args:
            inputs (array_like): A list of points. The shape must be '(n, 3)', where
                'n' means the number of points.
            num_frames (int): The number of frames taken.
            start_time (float, optional): A time to start detecting.
                Defaults to `0.0` in the configuration.
            exposure_time (float, optional): An exposure time.
                Defaults to `detector_exposure_time` in the configuration.
            full_output (bool, optional):
                True to return a dictionary containing optional outputs.

        Yields:
            Image: An image.
            dict: only returned if full_output == True
                A dictionary containing additional information.
                'expectation' is 2-dimensional ndarray.
                'true_data' is a dictionary containing pairs of a molecule ID and ndarray.
                The array has 7 elements, which are exposure time, photon state, X and Y in pixels,
                X and Y in meter-scale, and normalization.
                If 'photo_bleaching' is active, 'fluorescence_states' contains pairs of a molecule
                ID and remaining photon budget.
        """
        data = self.__format_inputs(inputs)
        base = self.base()
        for (camera, infodict) in base.generate_frames(
                data, num_frames, start_time=start_time, exposure_time=exposure_time, rng=self.__rng):
            img = scopyon.image.Image(camera[:, :, 1])
            if full_output:
                infodict.update(dict(expectation=camera[:, :, 0]))
                yield img, infodict
            else:
                yield img

def create_simulator(config=None, method=None, rng=None):
    """Return a simulator.

    Args:
        config (Configuration, optional): Configurations.
        method (str, optional): A name of method used.
            The default is None (config.default).
        rng (numpy.RandomState, optional): A random number generator.
            The default is None.

    Returns:
        EPIFMSimulator: A simulator
    """
    if method is None:
        method = config.get('method', 'default').lower()
    else:
        method = method

    simulator_type = config[method].type.lower()
    if simulator_type == 'epifm':
        return EPIFMSimulator(config=config, method=method, rng=rng)
    else:
        raise ValueError(f"An unknown type [{simulator_type}] was given.")

def form_image(
        inputs, start_time=0.0, exposure_time=None, *,
        method=None, config=None, rng=None, full_output=False):
    """Form image.

    Args:
        inputs (array_like): A list of points. The shape must be '(n, 3)',
            where 'n' means the number of points.
        start_time (float, optional): A time to start detecting.
            Defaults to 0.0.
        exposure_time (float, optional): An exposure time.
            Defaults to `detector.exposure_time` in the configuration.
        method (str, optional): A name of method used.
            The default is None ('default').
        config (Configuration, optional): Configurations.
            The default is None.
        rng (numpy.RandomState, optional): A random number generator.
            The default is None.
        full_output (bool, optional):
            True to return a dictionary containing optional outputs.

    Returns:
        Image: An image
        dict: only returned if full_output == True
            A dictionary containing additional information.
            'expectation' is 2-dimensional ndarray.
            'true_data' is a dictionary containing a pair of a molecule ID and ndarray.
            The array has 7 elements, which are exposure time, photon state, X and Y in pixels,
            X and Y in meter-scale, and normalization.
    """
    sim = create_simulator(config, method=method, rng=rng)
    return sim.form_image(inputs, start_time, exposure_time, full_output=full_output)

def generate_images(
        inputs, num_frames, start_time=0.0, exposure_time=None, *,
        method=None, config=None, rng=None, full_output=False):
    """Form images.

    Args:
        inputs (array_like): A list of points. The shape must be '(n, 3)',
            where 'n' means the number of points.
        num_frames (int): The number of frames taken.
        start_time (float, optional): A time to start detecting.
            Defaults to 0.0.
        exposure_time (float, optional): An exposure time.
            Defaults to `detector_exposure_time` in the configuration.
        method (str, optional): A name of method used.
            The default is None ('default').
        config (Configuration, optional): Configurations.
            The default is None.
        rng (numpy.RandomState, optional): A random number generator.
            The default is None.
        full_output (bool, optional):
            True to return a dictionary containing optional outputs.

    Yields:
        Image: An image.
        dict: only returned if full_output == True
            A dictionary containing additional information.
            'expectation' is 2-dimensional ndarray.
            'true_data' is a dictionary containing pairs of a molecule ID and ndarray.
            The array has 7 elements, which are exposure time, photon state, X and Y in pixels,
            X and Y in meter-scale, and normalization.
            If 'photo_bleaching' is active, 'fluorescence_states' contains pairs of a molecule
            ID and remaining photon budget.
    """
    sim = create_simulator(config, method=method, rng=rng)
    return sim.generate_images(inputs, num_frames, start_time, exposure_time, full_output=full_output)


if __name__ == "__main__":
    import sys
    filename = sys.argv[1]
    config = Configuration(filename)
    print(config.yaml)
    n = 10
    inputs = numpy.hstack((numpy.random.uniform(0, 5e-5, size=(n, 2)), numpy.zeros((n, 1)), ))
    print(form_image(inputs, config=config))



import unittest


class TestConfig(unittest.TestCase):

    def setUp(self):
        self.exposure_time = 0.1  # a default

    def tearDown(self):
        pass

    def test1(self):
        import scopyon.config

    def test2(self):
        from scopyon.config import DefaultConfiguration
        config = DefaultConfiguration()

    def test3(self):
        from scopyon.config import DefaultConfiguration
        config = DefaultConfiguration()
        self.assertAlmostEqual(config.default.detector.exposure_time, self.exposure_time)
        config.default.detector.exposure_time = 0.033
        self.assertAlmostEqual(config.default.detector.exposure_time, 0.033)

    def test4(self):
        from scopyon.config import DefaultConfiguration
        config = DefaultConfiguration()
        config.update("""
        default:
            detector:
                exposure_time: 0.033
        """)
        self.assertAlmostEqual(config.default.detector.exposure_time, 0.033)

        config.update("""
        default:
            detector:
                exposure_time:
                    value: 0.1
        """)
        self.assertAlmostEqual(config.default.detector.exposure_time, 0.1)

    def test4(self):
        from scopyon.config import DefaultConfiguration
        config = DefaultConfiguration()
        config.update("""
        default:
            detector:
                exposure_time:
                    value: 0.033
                    units: s
        """)
        self.assertAlmostEqual(config.default.detector.exposure_time, 0.033)

        config.update("""
        default:
            detector:
                exposure_time:
                    value: 100
                    units: ms
        """)
        self.assertAlmostEqual(config.default.detector.exposure_time, 0.1)

        config.default.detector.exposure_time = 0.033
        self.assertAlmostEqual(config.default.detector.exposure_time, 0.033)

    def test5(self):
        from scopyon.config import DefaultConfiguration
        from scopyon.constants import Q_
        config = DefaultConfiguration()
        config.default.detector.exposure_time = Q_(33, 'ms')
        self.assertAlmostEqual(config.default.detector.exposure_time, 0.033)

    def test5(self):
        from scopyon.config import DefaultConfiguration
        from scopyon.constants import Q_
        config = DefaultConfiguration()

        from pint.errors import DimensionalityError
        with self.assertRaises(DimensionalityError):
            config.default.detector.exposure_time = Q_(33, 'm')


if __name__ == '__main__':
    unittest.main()



import warnings
import numbers
import collections.abc

import numpy
import numpy.random

from logging import getLogger
_log = getLogger(__name__)

__all__ = ["sample_inputs"]


def sample_points(rng, N=None, conc=None, lower=None, upper=None, start=0, ndim=3):
    """Generate points distributed uniformly.

    Args:
        rng (numpy.RandomState, optional): A random number generator.
        N (int or list, optional): The number of points to be generated.
        conc (float or list, optional): The concentration of points.
            Either one of `N` or `conc` must be given.
        upper (Number or array, optional): An upper limit of the position. Defaults to 1.
        lower (Number or array, optional): A lower limit of the position. Defaults to 0.
        start (int, optional): The first index. Defaults to 0.
        ndim (int, optional) The number of dimensions. Defaults to 3.

    Returns:
        A pair of an array and the last ID.

        An array of points. Each point consists of a coordinate, an index,
        p_state (defaults to 1) and cyc_id (defaults to `inf`).

        The last ID. The sum of `start` and the number of points generated.

    """
    if N is None and conc is None:
        raise ValueError('Either one of N or conc must be given.')

    _log.info('sample_points: N={}, conc={}, lower={}, upper={}, start={}, ndim={}.'.format(
        N, conc, lower, upper, start, ndim))

    lower = (numpy.ones(ndim) * lower if isinstance(lower, numbers.Number)
        else numpy.array(lower) if lower is not None
        else numpy.zeros(ndim))
    upper = (numpy.ones(ndim) * upper if isinstance(upper, numbers.Number)
        else numpy.array(upper) if upper is not None
        else numpy.ones(ndim))

    if len(lower) != ndim or len(upper) != ndim:
        raise ValueError(
            "The wrong size of limits was given [{},{} != {}].".format(len(lower), len(upper), ndim))

    for dim in range(ndim):
        if lower[dim] > upper[dim]:
            lower[dim], upper[dim] = upper[dim], lower[dim]

    _log.debug('lower was set to {}.'.format(lower))
    _log.debug('upper was set to {}.'.format(upper))

    if N is None:
        lengths = upper - lower
        size = numpy.prod(lengths[lengths != 0])
        if isinstance(conc, collections.abc.Iterable):
            N_list = [rng.poisson(size * conc_) for conc_ in conc]
        else:
            N_list = [rng.poisson(size * conc)]
    elif not isinstance(N, collections.abc.Iterable):
        N_list = [N]
    else:
        N_list = N
    N = sum(N_list)
    _log.debug('{} points would be distributed {}.'.format(N, list(N_list)))

    if N <= 0:
        return numpy.array([]), start

    ret = numpy.zeros((N, ndim + 2))
    for dim in range(ndim):
        if lower[dim] < upper[dim]:
            ret[: , dim] = rng.uniform(lower[dim], upper[dim], N)
        else:
            ret[: , dim] = lower[dim]
    ret[: , ndim + 0] = numpy.arange(start, start + N)  # Molecule ID
    ret[: , ndim + 1] = 1.0  # Photon state
    return ret, start + N

def move_points(rng, points, D, dt, ndim=3):
    """
    Move points based on the given diffusion constant.

    Args:
        rng (numpy.RandomState, optional): A random number generator.
        points (array): Points.
        D (float or array-like): Diffusion constants.
            It is a constant or an array contains constants along with each axis.
        dt (float): A step interval.
        ndim (int, optional) The number of dimensions. Defaults to 3.

    Returns:
        array: An array of points updated.

    """
    _log.info('move_points: D={}, dt={}, ndim={}.'.format(D, dt, ndim))
    _log.debug('{} points are given.'.format(len(points)))

    if D is None:
        D = numpy.zeros(ndim)
    elif not isinstance(D, collections.abc.Iterable):
        D = numpy.ones(ndim) * D
    else:
        D = numpy.asarray(D)
        assert len(D) == ndim

    ret = points.copy()
    for i in range(len(ret)):
        for dim in range(ndim):
            ret[i, dim] += rng.normal(0.0, numpy.sqrt(2 * D[dim] * dt))
    return ret

def sample_inputs(t, *, N=None, conc=None, lower=None, upper=None, D=None, ndim=3, rng=None):
    """Generate the input data.

    Args:
        t (arraylike): The time points.
        N (int or list, optional): The number of points to be generated.
        conc (float or list, optional): The concentration of points.
            Either one of `N` or `conc` must be given.
        upper (Number or array, optional): An upper limit of the position. Defaults to 1.
        lower (Number or array, optional): A lower limit of the position. Defaults to 0.
        D (float or array-like, optional): Diffusion constants.
            It is a constant or an array contains constants along with each axis.
            Defaults to 0.0.
        ndim (int, optional) The number of dimensions. Defaults to 3.
        rng (numpy.RandomState, optional): A random number generator.

    Returns:
        A pair of an array and the last ID.

        An array of points. Each point consists of a coordinate, an index,
        p_state (defaults to 1) and cyc_id (defaults to `inf`).

        The last ID. The sum of `start` and the number of points generated.

    """
    if rng is None:
        warnings.warn('A random number generator [rng] is not given.')
        rng = numpy.random.RandomState()

    t = sorted(t)
    points, _ = sample_points(rng, N=N, conc=conc, lower=lower, upper=upper, ndim=ndim)
    tcurrent = t[0]
    inputs = []
    for tnext in t:
        if tnext > tcurrent:
            points = move_points(rng, points, D=D, dt=(tnext - tcurrent), ndim=ndim)
            tcurrent = tnext
        else:
            points = points.copy()
        inputs.append((tcurrent, points))
    return inputs

# def attempt_reactions(rng, points, dt, transitions=None, degradation=None, synthesis=None, lower=None, upper=None, start=0, ndim=3):
#     """
#     Apply first order reactions.
# 
#     Args:
#         rng (numpy.RandomState): A random number generator.
#         points (array): Points.
#         dt (float): A step interval.
#         transitions (matrix, optional): A matrix of kinetic rates.
#             k[i, j] means the first order kinetic rate for the transition from i-th to j-th.
#         degradation (array, optional): An array of degradation rates.
#             The probability that the reaction happens durint the step is given as `1 - exp(-k * dt)`.
#         synthesis (array, optional): An array of synthesis rates per unit volume.
#         upper (Number or array, optional): An upper limit of the position. Defaults to 1.
#         lower (Number or array, optional): A lower limit of the position. Defaults to 0.
#         start (int, optional): The first index. Defaults to 0.
#         ndim (int, optional) The number of dimensions. Defaults to 3.
# 
#     Returns:
#         array: An array of points updated.
#         int: The sum of `start` and the number of points generated.
# 
#     """
#     _log.info('attempt_reactions: dt={}.'.format(dt))
#     _log.debug('{} points are given.'.format(len(points)))
# 
#     ndim = 3
# 
#     ret = points.copy()
# 
#     if transitions is not None:
#         transitions = numpy.asarray(transitions)
#         Pacc = 1.0 - numpy.exp(-transitions * dt)
#         Pacc = numpy.add.accumulate(Pacc, 1)
# 
#         if any(Pacc[:, -1] > 1.0):
#             raise ValueError('The acceptance probability exceeds 1.')
# 
#         n, m = Pacc.shape
# 
#         rnd = rng.uniform(size=len(ret))
#         cnt = 0
#         for i, fluorophore_id in enumerate(ret[: , ndim + 2]):
#             fluorophore_id = int(fluorophore_id)
#             if not (0 <= fluorophore_id < n):
#                 raise ValueError(
#                     "The invalid fluorophore id '{}' was given.".format(fluorophore_id))
#             j = numpy.searchsorted(Pacc[fluorophore_id], rnd[i])
#             if j != m:
#                 ret[i, ndim + 2] = j
#                 cnt += 1
#         _log.debug('{} transitions occur.'.format(cnt))
# 
#     if degradation is not None:
#         degradation = numpy.asarray(degradation)
#         Pacc = 1.0 - numpy.exp(-degradation * dt)
#         Pacc = numpy.add.accumulate(Pacc)
# 
#         rnd = rng.uniform(size=len(ret))
#         surv = numpy.array([rnd[i] > Pacc[int(fluorophore_id)] for i, fluorophore_id in enumerate(ret[: , ndim + 2])])
#         ret = ret[surv]
#         _log.debug('{} points were degradated.'.format(len(ret) - len(points)))
# 
#     if synthesis is not None:
#         synthesis = numpy.asarray(synthesis)
#         ret_, start = sample_points(rng, conc=synthesis * dt, lower=lower, upper=upper, start=start)
#         if len(ret_) > 0:
#             ret = numpy.vstack((ret, ret_))
#             _log.debug('{} points were born.'.format(len(ret_)))
# 
#     return ret, start
# 
# 
# if __name__ == "__main__":
#     from .samples import sample_points, move_points, attempt_reactions
# 
# 
#     rng = numpy.random.RandomState(0)
#     lengths = numpy.array([2220.872e-9, 50033.2e-9, 50060e-9])
#     dt = 33e-3
#     N1 = 120
#     N2 = 60
#     D1 = 0.1e-6
#     D2 = 0.01e-6
#     k12 = k21 = -numpy.log(1.0 - 0.2) / dt
#     kd = k21
#     size = numpy.multiply.reduce(lengths[lengths != 0])
#     ks = N2 / size * kd
# 
#     points, start = sample_points(rng, upper=lengths, N=[N1, N2])
#     t = 0.0
#     print(points)
#     print(points[: , 3])
#     print(points[: , 5])
#     print(points.T[0].min(), points.T[0].max(), lengths[0])
#     print(points.T[1].min(), points.T[1].max(), lengths[1])
#     print(points.T[2].min(), points.T[2].max(), lengths[2])
# 
#     points = move_points(rng, points, [D1, D2], dt, lengths)
#     points, start = attempt_reactions(rng, points, dt, transitions=[[0.0, k12], [k21, 0.0]], degradation=[0.0, kd], synthesis=[ks, 0.0], upper=lengths, start=start)
#     t += dt
#     print(points)
#     print(points[: , 3])
#     print(points[: , 5])
#     print(points.T[0].min(), points.T[0].max(), lengths[0])
#     print(points.T[1].min(), points.T[1].max(), lengths[1])
#     print(points.T[2].min(), points.T[2].max(), lengths[2])
# 
#     # points, start = sample_points(rng, upper=lengths, N=[N1, N2])
#     # t = 0.0
#     # while t < dt * 100:
#     #     points = move_points(rng, points, [D1, D2], dt, lengths)
#     #     points, start = attempt_reactions(rng, points, dt, transitions=[[0.0, k12], [k21, 0.0]], degradation=[0.0, kd], synthesis=[ks, 0.0], upper=lengths, start=start)
#     #     t += dt
#     #     print(t, sum(points[: , 5] == 0), sum(points[: , 5] == 1))



import pint


__all__ = ['ureg', 'Q_', 'ABOGADROS_NUMBER', 'N_A', 'hc']

ureg = pint.UnitRegistry()
Q_ = ureg.Quantity

# Avogadro constant
# AVOGADROS_NUMBER = 6.022140857e+23
AVOGADROS_NUMBER = (1.0 * ureg.avogadro_number).to('dimensionless').magnitude
N_A = AVOGADROS_NUMBER

# (plank const) * (speed of light) [joules meter]
# hc = 6.62607015e-34 * 299792458 = 1.9864458571489289e-25
hc = (1.0 * ureg.planck_constant * ureg.speed_of_light).to(ureg.joule * ureg.meter).magnitude



import os.path
import numpy

from logging import getLogger
_log = getLogger(__name__)

__all__ = ["Image", "Video"]

from scopyon import _plotly, _matplotlib

class Image(object):

    PLOTTING = _plotly

    @staticmethod
    def load(file):
        """Load an image.

        Args:
            file (str): A file path.

        Returns:
            Image: an image object
        """
        assert isinstance(file, str)
        filename = file
        _, ext = os.path.splitext(filename)
        if ext.lower() == '.npy':
            assert len(kwargs) == 0  # No option is allowed
            return Image(numpy.load(filename))
        elif ext.lower() == '.csv':
            return Image(numpy.loadtxt(filename))

        #XXX: Expected to be an image format
        import matplotlib.pyplot as plt
        return Image(plt.imread(filename))

    @staticmethod
    def RGB(red=None, green=None, blue=None):
        """Make an image from RGB arrays.

        Args:
            red (2d ndarray or Image): RGB channel
            green (2d ndarray or Image): RGB channel
            blue (2d ndarray or Image): RGB channel

        Returns:
            Image: an image object
        """
        shape, dtype = None, None
        for channel in (red, green, blue):
            if channel is None:
                pass
            elif shape is None:
                shape = channel.shape
                dtype = channel.dtype
            else:
                assert shape == channel.shape and dtype == channel.dtype
        assert shape is not None
        rgb = numpy.zeros((shape[0], shape[1], 3), dtype=dtype)
        for i, channel in enumerate((red, green, blue)):
            if isinstance(channel, Image):
                rgb[:, :, i] = channel.as_array()
            elif isinstance(channel, numpy.ndarray):
                rgb[:, :, i] = channel
            else:
                assert channel is None
        return Image(rgb)

    def __init__(self, data):
        """A wrapper for an image.

        Args:
            data (2d or 3d ndarray): An image data.
        """
        assert data.ndim == 2 or (data.ndim == 3 and data.shape[2] == 3)
        self.__data = data

    def as_array(self):
        return self.__data

    @property
    def dtype(self):
        return self.__data.dtype

    @property
    def ndim(self):
        return self.__data.ndim

    @property
    def size(self):
        return self.__data.size

    @property
    def shape(self):
        return self.__data.shape

    @staticmethod
    def __as_8bit(data, cmin=None, cmax=None, low=None, high=None):
        """Same as scipy.misc.bytescale

        Args:
            data (array): data
            cmin (float, optional): Defaults to the minimum in data.
            cmax (float, optional): Defaults to the maximum in data.
            low (float, optional): Defaults to 0
            high (float, optional): Defaults to 255.

        Returns:
            array: 8bit data
        """
        cmin = data.min() if cmin is None else cmin
        cmax = data.max() if cmax is None else cmax
        low = 0 if low is None else low
        high = 255 if high is None else high

        cscale = cmax - cmin
        if cscale == 0.0:
            return numpy.ones(data.shape, dtype=numpy.uint8) * low
        scale = float(high - low) / cscale
        bytedata = (data - cmin) * scale + low
        bytedata = (bytedata.clip(low, high) + 0.5).astype(numpy.uint8)
        return bytedata

    def as_8bit(self, cmin=None, cmax=None, low=None, high=None):
        """Same as scipy.misc.bytescale

        Args:
            cmin (float, optional): Defaults to the minimum in data.
            cmax (float, optional): Defaults to the maximum in data.
            low (float, optional): Defaults to 0
            high (float, optional): Defaults to 255.

        Returns:
            Image: an Image object with 8bit data.
        """
        if self.dtype is numpy.uint8:
            data = self.__data.copy()
        elif self.ndim == 2:
            data = self.__as_8bit(
                    self.__data, cmin=cmin, cmax=cmax, low=low, high=high)
        elif self.ndim == 3:
            assert self.__data.shape[2] == 3
            data = numpy.zeros(self.__data.shape, dtype=numpy.uint8)
            for i in range(3):
                data[:, :, i] = self.__as_8bit(
                        self.__data[:, :, i], cmin=cmin, cmax=cmax, low=low, high=high)
        else:
            assert False
        return Image(data)

    def save(self, filename, **kwargs):
        """Save the 8-bit image.

        Note:
            Requires `pillow` to save in the image format.
            See also `scopyon.Image.savefig`.

        Args:
            filename (str): An output file name.
                `.npy`, `.csv` or image formats are accepted.
        """
        _, ext = os.path.splitext(filename)
        if ext.lower() == '.npy':
            assert len(kwargs) == 0  # No option is allowed
            numpy.save(filename, self.__data)
        elif ext.lower() == '.csv':
            assert len(kwargs) == 0  # No option is allowed
            numpy.savetxt(filename, self.__data)
        else:
            #XXX: Expected to be an image format
            self.savefig(filename, self.as_8bit().as_array(), **kwargs)

    @staticmethod
    def savefig(filename, img, shapes=None):
        """ Save figure.

        Args:
            filename (str): An output file path.
            img (ndarray): An image data to be shown.
            shapes (list, optional): A list of shapes.
                shape is a dictionary consisting of `x` (row), `y` (column), `sigma` and `color`.
                `sigma` is a half size of the box (square).
        """
        import PIL.Image
        from PIL import ImageDraw

        img = PIL.Image.fromarray(img)

        if shapes is not None:
            img = img.convert('RGB')
            d = ImageDraw.Draw(img)
            for shape in shapes:
                row, column, sigma, c = shape['x'], shape['y'], shape['sigma'], shape['color']
                x, y = column, row  # imshow
                d.rectangle([(x - sigma, y - sigma), (x + sigma, y + sigma)], outline=c, width=1)

        img.save(filename)
        # import matplotlib.pyplot as plt
        # plt.ioff()
        # _, ax = plt.subplots()
        # if img.ndim == 2:
        #     plt.imshow(img, interpolation='none', cmap='gray')
        # elif img.ndim == 3:
        #     plt.imshow(img, interpolation='none')
        # else:
        #     raise ValueError("'img' has wrong dimension.")
        # if shapes is not None:
        #     for shape in shapes:
        #         ax.add_patch(__get_shape(shape))
        # plt.axis('off')
        # plt.tick_params(labelbottom=False, labelleft=False, labelright=False, labeltop=False)
        # plt.savefig(filename, bbox_inches='tight', pad_inches=0)
        # plt.clf()
        # plt.ion()

    def show(self, **kwargs):
        """Plot the 8-bit image.

        Note:
            Requires `plotly` optionally.
            See also `scopyon._plotly.show`, `scopyon._matplotlib.show`.

        """
        try:
            self.PLOTTING.show(self.as_8bit().as_array(), **kwargs)
        except ImportError:
            _matplotlib.show(self.as_8bit().as_array(), **kwargs)

    def _ipython_display_(self):
        """
        Displays the object as a side effect.
        https://ipython.readthedocs.io/en/stable/config/integrating.html
        """
        self.show()

class Video:

    @staticmethod
    def save(filename, imgs, interval=100, dpi=None, cmin=None, cmax=None, low=None, high=None):
        """Make a video from images.

        Note:
            Requires `matplotlib`.

        Args:
            filename (str): An output file name.
            imgs (list): A list of Images.
            interval (int, optional): An interval between frames given in millisecond.
            dpi (int, optional): dpi.
        """
        import matplotlib.pyplot as plt
        from matplotlib.animation import FuncAnimation

        def animate(frame, image, imgs, cmin, cmax, low, high):
            image.set_array(imgs[frame].as_8bit(cmin=cmin, cmax=cmax, low=low, high=high).as_array())
            return image

        if cmin is None:
            cmin = min([img.as_array().min() for img in imgs])
        if cmax is None:
            cmax = max([img.as_array().max() for img in imgs])

        plt.ioff()
        fig, ax = plt.subplots(1, figsize=(1, 1))
        fig.subplots_adjust(0, 0, 1, 1)
        ax.axis("off")
        image = ax.imshow(imgs[0].as_8bit(cmin=cmin, cmax=cmax, low=low, high=high).as_array(), cmap='gray')
        animation = FuncAnimation(
            fig, animate, numpy.arange(len(imgs)), fargs=(image, imgs, cmin, cmax, low, high), interval=interval)
        dpi = dpi or max(imgs[0].shape)
        animation.save(filename, dpi=dpi)
        plt.clf()
        plt.ion()



import numpy
import scipy.optimize

from logging import getLogger
_log = getLogger(__name__)

__all__ = ["blob_detection", "spot_detection"]


def blob_detection(data, min_sigma=1, max_sigma=50, num_sigma=10, threshold=0.2, overlap=0.5):
    """Finds blobs in the given image.
    See also http://scikit-image.org/docs/dev/api/skimage.feature.html#skimage.feature.blob_log

    Note:
        Requires `scikit-image`.

    Args:
        data (ndarray): An image data.
        min_sigma (float, optional): The minimum standard deviation.
            Keep this low to detect smaller blobs. Defaults to 1.
        max_sigma (float, optional): The maximum standard deviation.
            Keep this high to detect larger blobs. Defaults to 50.
        num_sigma (int, optional): The number of intermediate values between `min_sigma` and `max_sigma`.
            Defaults to 10.
        threshold (float, optional): The absolute lower bound for scale space maxima.
            Reduce this to detect blobs with less intensities.
        overlap (float, optional): A value between 0 and 1.

    Returns:
        ndarray: Blobs detected.
            Each row represents coordinates and the standard deviation, `(x, y, r)`.
    """
    try:
        from skimage.feature import blob_log
    except ImportError:
        raise ImportError("No module named 'skimage'. 'spot_detection' requires 'scikit-image'")

    ## Laplacian of Gaussian
    blobs = blob_log(
        data, min_sigma=min_sigma, max_sigma=max_sigma, num_sigma=num_sigma, threshold=threshold, overlap=overlap)
    blobs[: , 2] = blobs[: , 2] * numpy.sqrt(2)
    _log.info('{} blob(s) were detected'.format(len(blobs)))
    return blobs

def mean_background(roi):
    m, n = roi.shape
    left = roi[0, : -1].sum()
    right = roi[-1, : -1].sum()
    bottom = roi[: -1, 0].sum()
    top = roi[: -1, -1].sum()
    tot = left + right + bottom + top - (roi[0, 0] + roi[0, -1] + roi[-1, 0] + roi[-1, -1])
    a7 = tot / (2 * (m + n - 2))
    # return roi.astype(dtype=numpy.float64) - a7
    return a7

def planar_background(roi):
    m, n = roi.shape
    a7_init = mean_background(roi)
    a5_init = a6_init = 0.0
    def error_func(p):
        a5, a6, a7 = p
        left = roi[0, :] - (numpy.arange(n) * a6 + a7)
        right = roi[-1, :] - ((m - 1) * a5 + numpy.arange(n) * a6 + a7)
        bottom = roi[:, 0] - (numpy.arange(m) * a5 + a7)
        top = roi[:, -1] - (numpy.arange(m) * a5 + (n - 1) * a6 + a7)
        return numpy.concatenate([left, right, bottom, top])
        # return (left ** 2).sum() + (right ** 2).sum() + (bottom ** 2).sum() + (top ** 2).sum()
    res = scipy.optimize.least_squares(error_func, (a5_init, a6_init, a7_init))
    if res.success <= 0:
        return None
    return res.x

def background(roi):
    m, n = roi.shape
    a5, a6, a7 = planar_background(roi)
    bg = numpy.tile(numpy.arange(m)[:, numpy.newaxis], (1, n)) * a5 + numpy.tile(numpy.arange(n), (m, 1)) * a6 + a7
    return roi.astype(dtype=numpy.float64) - bg, bg.sum()

def gaussian(a1, a2, a3, a4):
    return lambda x, y: (
        a1 * numpy.exp(-((x - a2) ** 2 + (y - a3) ** 2) / a4))

def weighted_com(data):
    total = data.sum()
    X, Y = numpy.indices(data.shape)
    x = (X * data).sum() / total
    y = (Y * data).sum() / total
    return (x, y)

def fitgaussian(data, roi_size):
    a2_init, a3_init = weighted_com(data)
    a1_init = 255
    a4_init = roi_size / 2
    error_func = lambda p: numpy.ravel(gaussian(*p)(*numpy.indices(data.shape)) - data)
    res = scipy.optimize.least_squares(error_func, (a1_init, a2_init, a3_init, a4_init))
    if res.success <= 0:
        return None
    return res.x

def spot_detection(data, roi_size=6, blobs=None, **kwargs):
    """Finds spots in the given image.

    Note:
        Requires `scipy`.

    Args:
        data (ndarray): An image data.
        roi_size (float, optional): A default value of a half of the ROI size.
            Defaults to 6.
        blobs (ndarray, optional): Blobs. Defaults to `None`. See also `blob_detection`.

    Returns:
        ndarray: Spots detected.
            Each row represents center position x, y, intensity, background, height, and sigma,
            `(center_x, center_y, intensity, bg, height, sigma)`.
    """

    if blobs is None:
        blobs = blob_detection(data, **kwargs)

    spots = []
    for blob in blobs:
        x, y, _ = blob

        x0, x1 = int(x - roi_size), int(x + roi_size) + 1
        y0, y1 = int(y - roi_size), int(y + roi_size) + 1
        x0, x1 = max(0, x0), min(data.shape[0], x1)
        y0, y1 = max(0, y0), min(data.shape[1], y1)
        roi = data[x0: x1, y0: y1]

        if roi.sum() <= 0:
            _log.debug("spot_detection skip a blob due to the low signal.")
            continue

        roi_, bg = background(roi)
        if roi_ is None:
            _log.debug("spot_detection skip a blob due to the failure in background.")
            continue
        res = fitgaussian(roi_, roi_size)
        if res is None:
            _log.debug("spot_detection skip a blob due to the failure in fitgaussian.")
            continue

        (height, center_x, center_y, sigma) = res
        if not (0 <= center_x < roi.shape[0]
                and 0 <= center_y < roi.shape[1]):
            _log.debug("spot_detection skip a blob due to invalid parameters fitted.")
            continue

        intensity = gaussian(*res)(*numpy.indices(roi.shape)).sum()
        center_x += x0
        center_y += y0
        spots.append((center_x, center_y, intensity, bg, height, sigma))

    _log.info('{} spot(s) were detected'.format(len(spots)))
    spots = numpy.array(spots)
    return spots



#!/usr/bin/env python

from setuptools import setup
from setuptools.command.install import install
import os
import os.path
import sys
import glob


install_requires = [pkg_name.rstrip('\r\n') for pkg_name in open('requirements.txt').readlines()]

# version
here = os.path.dirname(os.path.abspath(__file__))
version = next(
    (line.split('=')[1].strip().replace("'", '')
     for line in open(os.path.join(here, 'scopyon', '__init__.py'))
     if line.startswith('__version__ = ')),
    '0.0.0dev')  ## default

def readme():
    with open('README.md') as f:
        return f.read()

class verify_version_command(install):
    description = "verify that the git tag matches the version"

    def run(self):
        tag = os.getenv('CIRCLE_TAG')
        if tag is not None and tag != 'v' + version:
            info = "Git tag: {0} does not match the version of this library: {1}".format(tag, version)
            sys.exit(info)

setup(
    name='scopyon',
    version=version,
    url='https://github.com/ecell/scopyon',
    description='Monte Carlo simulation toolkit for bioimaging systems',
    long_description=readme(),
    long_description_content_type='text/markdown',
    author='Masaki Watabe',
    author_email='masaki@riken.jp',
    maintainer='Kazunari Kaizu',
    maintainer_email='kaizu@riken.jp',
    packages=['scopyon', 'scopyon.analysis'],
    package_dir={'scopyon': 'scopyon', 'scopyon.analysis': 'scopyon/analysis'},
    package_data={'scopyon': ['scopyon.yaml', 'catalog/*.txt', 'catalog/*/*.csv']},
    data_files=[('examples', ['examples/tirf.py', 'examples/tirf_000.png', 'examples/twocolor.py', 'examples/twocolor_000.png', 'examples/bleaching.py'])],
    install_requires=install_requires,
    test_suite='test',
    license='BSD-3-Clause',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        # 'Topic :: Scientific/Engineering :: Physics',
        # 'Topic :: Scientific/Engineering :: Visualization',
        'License :: OSI Approved :: BSD License',
        ],
    python_requires='>=3',
    cmdclass={'verify': verify_version_command}
    )



from .base import *
from .config import *
from .image import *
from .sampling import *
from . import constants
from . import analysis

__all__ = [
    "EnvironSettings", "EPIFMSimulator",
    "form_image", "generate_images", "create_simulator",
    "Configuration", "DefaultConfiguration",
    "Image", "Video",
    "sample_inputs",
    "constants", "analysis"
    ]

__version__ = '1.0.0a4'
